"""
Inspired by https://github.com/python-trio/outcome

An outcome is similar to the builtin `concurrent.futures.Future`
or `asyncio.Future`, but without being tied to a particular task model.
"""
import abc

from cocotb.utils import remove_traceback_frames


def capture(fn, *args, **kwargs):
    """Obtain an `Outcome` representing the result of a function call"""
    try:
        return Value(fn(*args, **kwargs))
    except BaseException as e:
        e = remove_traceback_frames(e, ["capture"])
        return Error(e)


class Outcome(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def send(self, gen):
        """Send or throw this outcome into a generator"""

    @abc.abstractmethod
    def get(self, gen):
        """Get the value of this outcome, or throw its exception"""


class Value(Outcome):
    def __init__(self, value):
        self.value = value

    def send(self, gen):
        return gen.send(self.value)

    def get(self):
        return self.value

    def __repr__(self):
        return f"Value({self.value!r})"


class Error(Outcome):
    def __init__(self, error):
        self.error = error

    def send(self, gen):
        return gen.throw(self.error)

    def get(self):
        raise self.error

    def __repr__(self):
        return f"Error({self.error!r})"



# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Potential Ventures Ltd,
#       SolarFlare Communications Inc nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""A collections of triggers which a testbench can await."""

import abc
import inspect
import warnings
from collections.abc import Awaitable
from decimal import Decimal
from numbers import Real
from typing import Optional, Union

import cocotb
from cocotb import outcomes, simulator
from cocotb.log import SimLog
from cocotb.utils import (
    ParametrizedSingleton,
    get_sim_steps,
    get_time_from_sim_steps,
    lazy_property,
    remove_traceback_frames,
)


def _pointer_str(obj):
    """
    Get the memory address of *obj* as used in :meth:`object.__repr__`.

    This is equivalent to ``sprintf("%p", id(obj))``, but python does not
    support ``%p``.
    """
    full_repr = object.__repr__(obj)  # gives "<{type} object at {address}>"
    return full_repr.rsplit(" ", 1)[1][:-1]


class TriggerException(Exception):
    pass


class Trigger(Awaitable):
    """Base class to derive from."""

    # __dict__ is needed here for the `.log` lazy_property below to work.
    # The implementation of `_PyObject_GenericGetAttrWithDict` suggests that
    # despite its inclusion, __slots__ will overall give speed and memory
    # improvements:
    #  - the `__dict__` is not actually constructed until it's needed, and that
    #    only happens if the `.log` attribute is used, where performance
    #    concerns no longer matter.
    #  - Attribute setting and getting will still go through the slot machinery
    #    first, as "data descriptors" take priority over dict access
    __slots__ = ("primed", "__weakref__", "__dict__")

    def __init__(self):
        self.primed = False

    @lazy_property
    def log(self):
        return SimLog("cocotb.%s" % (type(self).__qualname__), id(self))

    @abc.abstractmethod
    def prime(self, callback):
        """Set a callback to be invoked when the trigger fires.

        The callback will be invoked with a single argument, `self`.

        Sub-classes must override this, but should end by calling the base class
        method.

        .. warning::
            Do not call this directly within a :term:`task`. It is intended to be used
            only by the scheduler.
        """
        self.primed = True

    def unprime(self):
        """Remove the callback, and perform cleanup if necessary.

        After being un-primed, a Trigger may be re-primed again in the future.
        Calling `unprime` multiple times is allowed, subsequent calls should be
        a no-op.

        Sub-classes may override this, but should end by calling the base class
        method.

        .. warning::
            Do not call this directly within a :term:`task`. It is intended to be used
            only by the scheduler.
        """
        self.primed = False

    def __del__(self):
        # Ensure if a trigger drops out of scope we remove any pending callbacks
        self.unprime()

    @property
    def _outcome(self):
        """The result that `await this_trigger` produces in a coroutine.

        The default is to produce the trigger itself, which is done for
        ease of use with :class:`~cocotb.triggers.First`.
        """
        return outcomes.Value(self)

    def __await__(self):
        # hand the trigger back to the scheduler trampoline
        return (yield self)


class PythonTrigger(Trigger):
    """Python triggers don't use GPI at all.

    For example: notification of coroutine completion.
    """


class GPITrigger(Trigger):
    """Base Trigger class for GPI triggers.

    Consumes simulation time.
    """

    __slots__ = ("cbhdl",)

    def __init__(self):
        Trigger.__init__(self)

        # Required to ensure documentation can build
        # if simulator is not None:
        #    self.cbhdl = simulator.create_callback(self)
        # else:
        self.cbhdl = None

    def unprime(self):
        """Disable a primed trigger, can be re-primed."""
        if self.cbhdl is not None:
            self.cbhdl.deregister()
        self.cbhdl = None
        Trigger.unprime(self)


class Timer(GPITrigger):
    """Fire after the specified simulation time period has elapsed."""

    round_mode: str = "error"

    def __init__(
        self,
        time: Union[Real, Decimal] = None,
        units: str = "step",
        *,
        round_mode: Optional[str] = None,
        time_ps: Union[Real, Decimal] = None
    ) -> None:
        """
        Args:
           time: The time value.

               .. versionchanged:: 1.5.0
                  Previously this argument was misleadingly called `time_ps`.

           units: One of
               ``'step'``, ``'fs'``, ``'ps'``, ``'ns'``, ``'us'``, ``'ms'``, ``'sec'``.
               When *units* is ``'step'``,
               the timestep is determined by the simulator (see :make:var:`COCOTB_HDL_TIMEPRECISION`).

        Examples:

            >>> await Timer(100, units='ps')

            The time can also be a ``float``:

            >>> await Timer(100e-9, units='sec')

            which is particularly convenient when working with frequencies:

            >>> freq = 10e6  # 10 MHz
            >>> await Timer(1 / freq, units='sec')

            Other builtin exact numeric types can be used too:

            >>> from fractions import Fraction
            >>> await Timer(Fraction(1, 10), units='ns')

            >>> from decimal import Decimal
            >>> await Timer(Decimal('100e-9'), units='sec')

            These are most useful when using computed durations while
            avoiding floating point inaccuracies.

        See Also:
            :func:`~cocotb.utils.get_sim_steps`

        Raises:
            TriggerException: If a negative value is passed for Timer setup.

        .. versionchanged:: 1.5
            Raise an exception when Timer uses a negative value as it is undefined behavior.
            Warn for 0 as this will cause erratic behavior in some simulators as well.

        .. versionchanged:: 1.5
            Support ``'step'`` as the *units* argument to mean "simulator time step".

        .. deprecated:: 1.5
            Using ``None`` as the *units* argument is deprecated, use ``'step'`` instead.

        .. versionchanged:: 1.6
            Support rounding modes.
        """
        GPITrigger.__init__(self)
        if time_ps is not None:
            if time is not None:
                raise TypeError(
                    "Gave argument to both the 'time' and deprecated 'time_ps' parameter"
                )
            time = time_ps
            warnings.warn(
                "The parameter name 'time_ps' has been renamed to 'time'. Please update your invocation.",
                DeprecationWarning,
                stacklevel=2,
            )
        else:
            if time is None:
                raise TypeError("Missing required argument 'time'")
        if time <= 0:
            if time == 0:
                warnings.warn(
                    "Timer setup with value 0, which might exhibit undefined behavior in some simulators",
                    category=RuntimeWarning,
                    stacklevel=2,
                )
            else:
                raise TriggerException("Timer value time_ps must not be negative")
        if units is None:
            warnings.warn(
                'Using units=None is deprecated, use units="step" instead.',
                DeprecationWarning,
                stacklevel=2,
            )
            units = "step"  # don't propagate deprecated value
        if round_mode is None:
            round_mode = type(self).round_mode
        self.sim_steps = get_sim_steps(time, units, round_mode=round_mode)

    def prime(self, callback):
        """Register for a timed callback."""
        if self.cbhdl is None:
            self.cbhdl = simulator.register_timed_callback(
                self.sim_steps, callback, self
            )
            if self.cbhdl is None:
                raise TriggerException("Unable set up %s Trigger" % (str(self)))
        GPITrigger.prime(self, callback)

    def __repr__(self):
        return "<{} of {:1.2f}ps at {}>".format(
            type(self).__qualname__,
            get_time_from_sim_steps(self.sim_steps, units="ps"),
            _pointer_str(self),
        )


# This is needed to make our custom metaclass work with abc.ABCMeta used in the
# `Trigger` base class.
class _ParameterizedSingletonAndABC(ParametrizedSingleton, abc.ABCMeta):
    pass


class ReadOnly(GPITrigger, metaclass=_ParameterizedSingletonAndABC):
    """Fires when the current simulation timestep moves to the read-only phase.

    The read-only phase is entered when the current timestep no longer has any further delta steps.
    This will be a point where all the signal values are stable as there are no more RTL events scheduled for the timestep.
    The simulator will not allow scheduling of more events in this timestep.
    Useful for monitors which need to wait for all processes to execute (both RTL and cocotb) to ensure sampled signal values are final.
    """

    __slots__ = ()

    @classmethod
    def __singleton_key__(cls):
        return None

    def __init__(self):
        GPITrigger.__init__(self)

    def prime(self, callback):
        if self.cbhdl is None:
            self.cbhdl = simulator.register_readonly_callback(callback, self)
            if self.cbhdl is None:
                raise TriggerException("Unable set up %s Trigger" % (str(self)))
        GPITrigger.prime(self, callback)

    def __repr__(self):
        return "{}()".format(type(self).__qualname__)


class ReadWrite(GPITrigger, metaclass=_ParameterizedSingletonAndABC):
    """Fires when the read-write portion of the simulation cycles is reached."""

    __slots__ = ()

    @classmethod
    def __singleton_key__(cls):
        return None

    def __init__(self):
        GPITrigger.__init__(self)

    def prime(self, callback):
        if self.cbhdl is None:
            self.cbhdl = simulator.register_rwsynch_callback(callback, self)
            if self.cbhdl is None:
                raise TriggerException("Unable set up %s Trigger" % (str(self)))
        GPITrigger.prime(self, callback)

    def __repr__(self):
        return "{}()".format(type(self).__qualname__)


class NextTimeStep(GPITrigger, metaclass=_ParameterizedSingletonAndABC):
    """Fires when the next time step is started."""

    __slots__ = ()

    @classmethod
    def __singleton_key__(cls):
        return None

    def __init__(self):
        GPITrigger.__init__(self)

    def prime(self, callback):
        if self.cbhdl is None:
            self.cbhdl = simulator.register_nextstep_callback(callback, self)
            if self.cbhdl is None:
                raise TriggerException("Unable set up %s Trigger" % (str(self)))
        GPITrigger.prime(self, callback)

    def __repr__(self):
        return "{}()".format(type(self).__qualname__)


class _EdgeBase(GPITrigger, metaclass=_ParameterizedSingletonAndABC):
    """Internal base class that fires on a given edge of a signal."""

    __slots__ = ("signal",)

    @classmethod
    @property
    def _edge_type(self):
        """The edge type, as understood by the C code. Must be set in sub-classes."""
        raise NotImplementedError

    @classmethod
    def __singleton_key__(cls, signal):
        return signal

    def __init__(self, signal):
        super().__init__()
        self.signal = signal

    def prime(self, callback):
        """Register notification of a value change via a callback"""
        if self.cbhdl is None:
            self.cbhdl = simulator.register_value_change_callback(
                self.signal._handle, callback, type(self)._edge_type, self
            )
            if self.cbhdl is None:
                raise TriggerException("Unable set up %s Trigger" % (str(self)))
        super().prime(callback)

    def __repr__(self):
        return "{}({!r})".format(type(self).__qualname__, self.signal)


class RisingEdge(_EdgeBase):
    """Fires on the rising edge of *signal*, on a transition from ``0`` to ``1``."""

    __slots__ = ()
    _edge_type = 1


class FallingEdge(_EdgeBase):
    """Fires on the falling edge of *signal*, on a transition from ``1`` to ``0``."""

    __slots__ = ()
    _edge_type = 2


class Edge(_EdgeBase):
    """Fires on any value change of *signal*."""

    __slots__ = ()
    _edge_type = 3


class _Event(PythonTrigger):
    """Unique instance used by the Event object.

    One created for each attempt to wait on the event so that the scheduler
    can maintain a dictionary of indexing each individual coroutine.

    FIXME: This will leak - need to use peers to ensure everything is removed
    """

    def __init__(self, parent):
        PythonTrigger.__init__(self)
        self.parent = parent

    def prime(self, callback):
        self._callback = callback
        self.parent._prime_trigger(self, callback)
        Trigger.prime(self, callback)

    def __call__(self):
        self._callback(self)

    def __repr__(self):
        return "<{!r}.wait() at {}>".format(self.parent, _pointer_str(self))


class Event:
    """Event to permit synchronization between two coroutines.

    Awaiting :meth:`wait()` from one coroutine will block the coroutine until
    :meth:`set()` is called somewhere else.
    """

    def __init__(self, name=None):
        self._pending = []
        self.name = name
        self.fired = False
        self.data = None

    def _prime_trigger(self, trigger, callback):
        self._pending.append(trigger)

    def set(self, data=None):
        """Wake up all coroutines blocked on this event."""
        self.fired = True
        self.data = data

        p = self._pending[:]

        self._pending = []

        for trigger in p:
            trigger()

    def wait(self):
        """Get a trigger which fires when another coroutine sets the event.

        If the event has already been set, the trigger will fire immediately.

        To reset the event (and enable the use of ``wait`` again),
        :meth:`clear` should be called.
        """
        if self.fired:
            return NullTrigger(name="{}.wait()".format(str(self)))
        return _Event(self)

    def clear(self):
        """Clear this event that has fired.

        Subsequent calls to :meth:`~cocotb.triggers.Event.wait` will block until
        :meth:`~cocotb.triggers.Event.set` is called again."""
        self.fired = False

    def is_set(self) -> bool:
        """Return true if event has been set"""
        return self.fired

    def __repr__(self):
        if self.name is None:
            fmt = "<{0} at {2}>"
        else:
            fmt = "<{0} for {1} at {2}>"
        return fmt.format(type(self).__qualname__, self.name, _pointer_str(self))


class _InternalEvent(PythonTrigger):
    """Event used internally for triggers that need cross-coroutine synchronization.

    This Event can only be waited on once, by a single coroutine.

    Provides transparent __repr__ pass-through to the Trigger using this event,
    providing a better debugging experience.
    """

    def __init__(self, parent):
        PythonTrigger.__init__(self)
        self.parent = parent
        self._callback = None
        self.fired = False
        self.data = None

    def prime(self, callback):
        if self._callback is not None:
            raise RuntimeError("This Trigger may only be awaited once")
        self._callback = callback
        Trigger.prime(self, callback)
        if self.fired:
            self._callback(self)

    def set(self, data=None):
        """Wake up coroutine blocked on this event."""
        self.fired = True
        self.data = data

        if self._callback is not None:
            self._callback(self)

    def is_set(self) -> bool:
        """Return true if event has been set."""
        return self.fired

    def __await__(self):
        if self.primed:
            raise RuntimeError("Only one coroutine may await this Trigger")
        # hand the trigger back to the scheduler trampoline
        return (yield self)

    def __repr__(self):
        return repr(self.parent)


class _Lock(PythonTrigger):
    """Unique instance used by the Lock object.

    One created for each attempt to acquire the Lock so that the scheduler
    can maintain a dictionary of indexing each individual coroutine.

    FIXME: This will leak - need to use peers to ensure everything is removed.
    """

    def __init__(self, parent):
        PythonTrigger.__init__(self)
        self.parent = parent

    def prime(self, callback):
        self._callback = callback
        self.parent._prime_trigger(self, callback)
        Trigger.prime(self, callback)

    def __call__(self):
        self._callback(self)

    def __repr__(self):
        return "<{!r}.acquire() at {}>".format(self.parent, _pointer_str(self))


class Lock:
    """Lock primitive (not re-entrant).

    This can be used as::

        await lock.acquire()
        try:
            # do some stuff
        finally:
            lock.release()

    .. versionchanged:: 1.4

        The lock can be used as an asynchronous context manager in an
        :keyword:`async with` statement::

            async with lock:
                # do some stuff
    """

    def __init__(self, name=None):
        self._pending_unprimed = []
        self._pending_primed = []
        self.name = name
        self.locked = False  #: ``True`` if the lock is held.

    def _prime_trigger(self, trigger, callback):
        self._pending_unprimed.remove(trigger)

        if not self.locked:
            self.locked = True
            callback(trigger)
        else:
            self._pending_primed.append(trigger)

    def acquire(self):
        """Produce a trigger which fires when the lock is acquired."""
        trig = _Lock(self)
        self._pending_unprimed.append(trig)
        return trig

    def release(self):
        """Release the lock."""
        if not self.locked:
            raise TriggerException(
                "Attempt to release an unacquired Lock %s" % (str(self))
            )

        self.locked = False

        # nobody waiting for this lock
        if not self._pending_primed:
            return

        trigger = self._pending_primed.pop(0)
        self.locked = True
        trigger()

    def __repr__(self):
        if self.name is None:
            fmt = "<{0} [{2} waiting] at {3}>"
        else:
            fmt = "<{0} for {1} [{2} waiting] at {3}>"
        return fmt.format(
            type(self).__qualname__,
            self.name,
            len(self._pending_primed),
            _pointer_str(self),
        )

    def __bool__(self):
        """Provide boolean of a Lock"""
        return self.locked

    async def __aenter__(self):
        return await self.acquire()

    async def __aexit__(self, exc_type, exc, tb):
        self.release()


class NullTrigger(Trigger):
    """Fires immediately.

    Primarily for internal scheduler use.
    """

    def __init__(self, name=None, outcome=None):
        super().__init__()
        self._callback = None
        self.name = name
        self.__outcome = outcome

    @property
    def _outcome(self):
        if self.__outcome is not None:
            return self.__outcome
        return super()._outcome

    def prime(self, callback):
        callback(self)

    def __repr__(self):
        if self.name is None:
            fmt = "<{0} at {2}>"
        else:
            fmt = "<{0} for {1} at {2}>"
        return fmt.format(type(self).__qualname__, self.name, _pointer_str(self))


class Join(PythonTrigger, metaclass=_ParameterizedSingletonAndABC):
    r"""Fires when a task completes.

    The result of blocking on the trigger can be used to get the coroutine
    result::

        async def coro_inner():
            await Timer(1, units='ns')
            return "Hello world"

        task = cocotb.start_soon(coro_inner())
        result = await Join(task)
        assert result == "Hello world"

    If the coroutine threw an exception, the :keyword:`await` will re-raise it.

    """
    __slots__ = ("_coroutine",)

    @classmethod
    def __singleton_key__(cls, coroutine):
        return coroutine

    def __init__(self, coroutine):
        super().__init__()
        self._coroutine = coroutine

    @property
    def _outcome(self):
        return self._coroutine._outcome

    @property
    def retval(self):
        """The return value of the joined coroutine.

        .. note::
            Typically there is no need to use this attribute - the
            following code samples are equivalent::

                forked = cocotb.start_soon(mycoro())
                j = Join(forked)
                await j
                result = j.retval

            ::

                forked = cocotb.start_soon(mycoro())
                result = await Join(forked)
        """
        return self._coroutine.result()

    def prime(self, callback):
        if self._coroutine.done():
            callback(self)
        else:
            super().prime(callback)

    def __repr__(self):
        return "{}({!s})".format(type(self).__qualname__, self._coroutine)


class Waitable(Awaitable):
    """
    Base class for trigger-like objects implemented using coroutines.

    This converts a `_wait` abstract method into a suitable `__await__`.
    """

    __slots__ = ()

    async def _wait(self):
        """
        Should be implemented by the sub-class. Called by `await self` to
        convert the waitable object into a coroutine.
        """
        raise NotImplementedError

    def __await__(self):
        return self._wait().__await__()


class _AggregateWaitable(Waitable):
    """
    Base class for Waitables that take mutiple triggers in their constructor
    """

    __slots__ = ("triggers",)

    def __init__(self, *triggers):
        self.triggers = triggers

        # Do some basic type-checking up front, rather than waiting until we
        # await them.
        allowed_types = (Trigger, Waitable, cocotb.decorators.Task)
        for trigger in self.triggers:
            if not isinstance(trigger, allowed_types):
                raise TypeError(
                    "All triggers must be instances of Trigger! Got: {}".format(
                        type(trigger).__qualname__
                    )
                )

    def __repr__(self):
        # no _pointer_str here, since this is not a trigger, so identity
        # doesn't matter.
        return "{}({})".format(
            type(self).__qualname__,
            ", ".join(
                repr(Join(t)) if isinstance(t, cocotb.decorators.Task) else repr(t)
                for t in self.triggers
            ),
        )


async def _wait_callback(trigger, callback):
    """
    Wait for a trigger, and call `callback` with the outcome of the await.
    """
    try:
        ret = outcomes.Value(await trigger)
    except BaseException as exc:
        # hide this from the traceback
        ret = outcomes.Error(remove_traceback_frames(exc, ["_wait_callback"]))
    callback(ret)


class Combine(_AggregateWaitable):
    """
    Fires when all of *triggers* have fired.

    Like most triggers, this simply returns itself.

    This is similar to Verilog's ``join``.
    """

    __slots__ = ()

    async def _wait(self):
        waiters = []
        e = _InternalEvent(self)
        triggers = list(self.triggers)

        # start a parallel task for each trigger
        for t in triggers:
            # t=t is needed for the closure to bind correctly
            def on_done(ret, t=t):
                triggers.remove(t)
                if not triggers:
                    e.set()
                ret.get()  # re-raise any exception

            waiters.append(cocotb.start_soon(_wait_callback(t, on_done)))

        # wait for the last waiter to complete
        await e
        return self


class First(_AggregateWaitable):
    """
    Fires when the first trigger in *triggers* fires.

    Returns the result of the trigger that fired.

    This is similar to Verilog's ``join_any``.

    .. note::
        The event loop is single threaded, so while events may be simultaneous
        in simulation time, they can never be simultaneous in real time.
        For this reason, the value of ``t_ret is t1`` in the following example
        is implementation-defined, and will vary by simulator::

            t1 = Timer(10, units='ps')
            t2 = Timer(10, units='ps')
            t_ret = await First(t1, t2)

    .. note::
        In the old-style :ref:`generator-based coroutines <yield-syntax>`, ``t = yield [a, b]`` was another spelling of
        ``t = yield First(a, b)``. This spelling is no longer available when using :keyword:`await`-based
        coroutines.
    """

    __slots__ = ()

    async def _wait(self):
        waiters = []
        e = _InternalEvent(self)
        completed = []
        # start a parallel task for each trigger
        for t in self.triggers:

            def on_done(ret):
                completed.append(ret)
                e.set()

            waiters.append(cocotb.start_soon(_wait_callback(t, on_done)))

        # wait for a waiter to complete
        await e

        # kill all the other waiters
        # TODO: Should this kill the coroutines behind any Join triggers?
        # Right now it does not.
        for w in waiters:
            w.kill()

        # These lines are the way they are to make tracebacks readable:
        #  - The comment helps the user understand why they are seeing the
        #    traceback, even if it is obvious top cocotb maintainers.
        #  - Using `NullTrigger` here instead of `result = completed[0].get()`
        #    means we avoid inserting an `outcome.get` frame in the traceback
        first_trigger = NullTrigger(outcome=completed[0])
        return await first_trigger  # the first of multiple triggers that fired


class ClockCycles(Waitable):
    """Fires after *num_cycles* transitions of *signal* from ``0`` to ``1``."""

    def __init__(self, signal, num_cycles, rising=True):
        """
        Args:
            signal: The signal to monitor.
            num_cycles (int): The number of cycles to count.
            rising (bool, optional): If ``True``, the default, count rising edges.
                Otherwise, count falling edges.
        """
        self.signal = signal
        self.num_cycles = num_cycles
        if rising is True:
            self._type = RisingEdge
        else:
            self._type = FallingEdge

    async def _wait(self):
        trigger = self._type(self.signal)
        for _ in range(self.num_cycles):
            await trigger
        return self

    def __repr__(self):
        # no _pointer_str here, since this is not a trigger, so identity
        # doesn't matter.
        if self._type is RisingEdge:
            fmt = "{}({!r}, {!r})"
        else:
            fmt = "{}({!r}, {!r}, rising=False)"
        return fmt.format(type(self).__qualname__, self.signal, self.num_cycles)


async def with_timeout(trigger, timeout_time, timeout_unit="step"):
    r"""
    Waits on triggers or coroutines, throws an exception if it waits longer than the given time.

    When a :term:`python:coroutine` is passed,
    the callee coroutine is started,
    the caller blocks until the callee completes,
    and the callee's result is returned to the caller.
    If timeout occurs, the callee is killed
    and :exc:`SimTimeoutError` is raised.

    When an unstarted :class:`~cocotb.coroutine`\ is passed,
    the callee coroutine is started,
    the caller blocks until the callee completes,
    and the callee's result is returned to the caller.
    If timeout occurs, the callee `continues to run`
    and :exc:`SimTimeoutError` is raised.

    When a :term:`task` is passed,
    the caller blocks until the callee completes
    and the callee's result is returned to the caller.
    If timeout occurs, the callee `continues to run`
    and :exc:`SimTimeoutError` is raised.

    If a :class:`~cocotb.triggers.Trigger` or :class:`~cocotb.triggers.Waitable` is passed,
    the caller blocks until the trigger fires,
    and the trigger is returned to the caller.
    If timeout occurs, the trigger is cancelled
    and :exc:`SimTimeoutError` is raised.

    Usage:

    .. code-block:: python

        await with_timeout(coro, 100, 'ns')
        await with_timeout(First(coro, event.wait()), 100, 'ns')

    Args:
        trigger (:class:`~cocotb.triggers.Trigger`, :class:`~cocotb.triggers.Waitable`, :class:`~cocotb.decorators.Task`, or :term:`python:coroutine`):
            A single object that could be right of an :keyword:`await` expression in cocotb.
        timeout_time (numbers.Real or decimal.Decimal):
            Simulation time duration before timeout occurs.
        timeout_unit (str, optional):
            Units of timeout_time, accepts any units that :class:`~cocotb.triggers.Timer` does.

    Returns:
        First trigger that completed if timeout did not occur.

    Raises:
        :exc:`SimTimeoutError`: If timeout occurs.

    .. versionadded:: 1.3

    .. deprecated:: 1.5
        Using ``None`` as the *timeout_unit* argument is deprecated, use ``'step'`` instead.

    .. versionchanged:: 1.7.0
        Support passing :term:`python:coroutine`\ s.
    """
    if timeout_unit is None:
        warnings.warn(
            'Using timeout_unit=None is deprecated, use timeout_unit="step" instead.',
            DeprecationWarning,
            stacklevel=2,
        )
        timeout_unit = "step"  # don't propagate deprecated value
    if inspect.iscoroutine(trigger):
        trigger = cocotb.start_soon(trigger)
        shielded = False
    else:
        shielded = True
    timeout_timer = cocotb.triggers.Timer(timeout_time, timeout_unit)
    res = await First(timeout_timer, trigger)
    if res is timeout_timer:
        if not shielded:
            trigger.kill()
        raise cocotb.result.SimTimeoutError
    else:
        return res



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

import abc
import os
import re
import shutil
import subprocess
import sys
import tempfile
import warnings
from contextlib import suppress
from typing import Dict, List, Mapping, Optional, Sequence, Type, Union
from xml.etree import cElementTree as ET

import cocotb.config

PathLike = Union["os.PathLike[str]", str]
Command = List[str]

warnings.warn(
    "Python runners and associated APIs are an experimental feature and subject to change.",
    UserWarning,
    stacklevel=2,
)

_magic_re = re.compile(r"([\\{}])")
_space_re = re.compile(r"([\s])", re.ASCII)


def as_tcl_value(value: str) -> str:
    # add '\' before special characters and spaces
    value = _magic_re.sub(r"\\\1", value)
    value = value.replace("\n", r"\n")
    value = _space_re.sub(r"\\\1", value)
    if value[0] == '"':
        value = "\\" + value

    return value


class Simulator(abc.ABC):
    def __init__(self) -> None:

        self.simulator_in_path()

        self.env: Dict[str, str] = {}

        # for running test() independently of build()
        self.build_dir = "sim_build"
        self.parameters = {}

    @abc.abstractmethod
    def simulator_in_path(self) -> None:
        """Check that the simulator executable exists in :envvar:`PATH`."""

        raise NotImplementedError()

    @abc.abstractmethod
    def check_toplevel_lang(self, toplevel_lang: Optional[str]) -> str:
        """Return *toplevel_lang* if supported by simulator, raise exception otherwise."""

        raise NotImplementedError()

    def set_env(self) -> None:
        """Set environment variables for sub-processes."""

        for e in os.environ:
            self.env[e] = os.environ[e]

        if "LIBPYTHON_LOC" not in self.env:
            self.env["LIBPYTHON_LOC"] = cocotb._vendor.find_libpython.find_libpython()

        self.env["PATH"] += os.pathsep + cocotb.config.libs_dir

        self.env["PYTHONPATH"] = os.pathsep.join(sys.path)
        for path in self.python_search:
            self.env["PYTHONPATH"] += os.pathsep + str(path)

        self.env["PYTHONHOME"] = sys.prefix

        self.env["TOPLEVEL"] = self.sim_toplevel
        self.env["MODULE"] = self.module

    @abc.abstractmethod
    def build_command(self) -> Sequence[Command]:
        """Return command to build the HDL sources."""

        raise NotImplementedError()

    @abc.abstractmethod
    def test_command(self) -> Sequence[Command]:
        """Return command to run a test."""

        raise NotImplementedError()

    def build(
        self,
        library_name: str = "work",
        verilog_sources: Sequence[PathLike] = [],
        vhdl_sources: Sequence[PathLike] = [],
        includes: Sequence[PathLike] = [],
        defines: Sequence[str] = [],
        parameters: Mapping[str, object] = {},
        extra_args: Sequence[str] = [],
        toplevel: Optional[str] = None,
        always: bool = False,
        build_dir: PathLike = "sim_build",
    ) -> None:
        """Build the HDL sources."""

        self.build_dir = os.path.abspath(build_dir)
        os.makedirs(self.build_dir, exist_ok=True)

        # note: to avoid mutating argument defaults, we ensure that no value
        # is written without a copy. This is much more concise and leads to
        # a better docstring than using `None` as a default in the parameters
        # list.
        self.library_name = library_name
        self.verilog_sources = get_abs_paths(verilog_sources)
        self.vhdl_sources = get_abs_paths(vhdl_sources)
        self.includes = get_abs_paths(includes)
        self.defines = list(defines)
        self.parameters = dict(parameters)
        self.compile_args = list(extra_args)
        self.always = always
        self.hdl_toplevel = toplevel

        for e in os.environ:
            self.env[e] = os.environ[e]

        cmds = self.build_command()
        self.execute(cmds, cwd=self.build_dir)

    def test(
        self,
        py_module: Union[str, Sequence[str]],
        toplevel: str,
        toplevel_lang: Optional[str] = None,
        testcase: Optional[str] = None,
        seed: Optional[Union[str, int]] = None,
        python_search: Sequence[PathLike] = [],
        extra_args: Sequence[str] = [],
        plus_args: Sequence[str] = [],
        extra_env: Mapping[str, str] = {},
        waves: Optional[bool] = None,
        gui: Optional[bool] = None,
        parameters: Mapping[str, object] = None,
        build_dir: Optional[PathLike] = None,
        sim_dir: Optional[PathLike] = None,
    ) -> PathLike:
        """Run a test."""

        __tracebackhide__ = True  # Hide the traceback when using pytest

        if build_dir is not None:
            self.build_dir = build_dir

        if parameters is not None:
            self.parameters = dict(parameters)

        if sim_dir is None:
            self.sim_dir = self.build_dir
        else:
            self.sim_dir = os.path.abspath(sim_dir)

        if isinstance(py_module, str):
            self.module = py_module
        else:
            self.module = ",".join(py_module)

        # note: to avoid mutating argument defaults, we ensure that no value
        # is written without a copy. This is much more concise and leads to
        # a better docstring than using `None` as a default in the parameters
        # list.
        self.python_search = list(python_search)
        self.sim_toplevel = toplevel
        self.toplevel_lang = self.check_toplevel_lang(toplevel_lang)
        self.sim_args = list(extra_args)
        self.plus_args = list(plus_args)
        self.env = dict(extra_env)

        if testcase is not None:
            self.env["TESTCASE"] = testcase

        if seed is not None:
            self.env["RANDOM_SEED"] = str(seed)

        if waves is None:
            self.waves = bool(int(os.getenv("COCOTB_WAVES", 0)))
        else:
            self.waves = bool(waves)

        if gui is None:
            self.gui = bool(int(os.getenv("COCOTB_GUI", 0)))
        else:
            self.gui = bool(gui)

        # When using pytest, use test name as result file name
        pytest_current_test = os.getenv("PYTEST_CURRENT_TEST", "")
        if pytest_current_test:
            self.current_test_name = pytest_current_test.split(":")[-1].split(" ")[0]
            results_xml_name = f"{self.current_test_name}.results.xml"
        else:
            self.current_test_name = "test"
            results_xml_name = "results.xml"

        results_xml_file = os.getenv(
            "COCOTB_RESULTS_FILE", os.path.join(self.build_dir, results_xml_name)
        )

        self.env["COCOTB_RESULTS_FILE"] = results_xml_file

        with suppress(OSError):
            os.remove(results_xml_file)

        cmds = self.test_command()
        self.set_env()
        self.execute(cmds, cwd=self.sim_dir)

        check_results_file(results_xml_file)

        print(f"INFO: Results file: {results_xml_file}")
        return results_xml_file

    @abc.abstractmethod
    def get_include_options(self, includes: Sequence[str]) -> List[str]:
        """Return simulator options setting directories searched for Verilog include files."""

        raise NotImplementedError()

    @abc.abstractmethod
    def get_define_options(self, defines: Sequence[str]) -> List[str]:
        """Return simulator options defining macros."""

        raise NotImplementedError()

    @abc.abstractmethod
    def get_parameter_options(self, parameters: Mapping[str, object]) -> List[str]:
        """Return simulator options for module parameters/generics."""

        raise NotImplementedError()

    def execute(self, cmds: Sequence[Command], cwd: PathLike) -> None:

        __tracebackhide__ = True  # Hide the traceback when using PyTest.

        for cmd in cmds:
            print(
                "INFO: Running command: "
                + ' "'.join(cmd)
                + '" in directory:"'
                + str(cwd)
                + '"'
            )

            # TODO: create a thread to handle stderr and log as error?
            # TODO: log forwarding

            process = subprocess.run(cmd, cwd=cwd, env=self.env)

            if process.returncode != 0:
                raise SystemExit(
                    f"Process {process.args[0]!r} terminated with error {process.returncode}"
                )


def check_results_file(results_xml_file: PathLike) -> None:
    """Check whether cocotb result file exists and contains failed tests."""

    __tracebackhide__ = True  # Hide the traceback when using PyTest.

    results_file_exist = os.path.isfile(results_xml_file)
    if not results_file_exist:
        raise SystemExit(
            "ERROR: Simulation terminated abnormally. Results file not found."
        )

    failed = 0

    tree = ET.parse(results_xml_file)
    for ts in tree.iter("testsuite"):
        for tc in ts.iter("testcase"):
            for _ in tc.iter("failure"):
                failed += 1

    if failed:
        raise SystemExit(f"ERROR: Failed {failed} tests.")


def outdated(output: PathLike, dependencies: Sequence[PathLike]) -> bool:
    """Check if source files are newer than output."""

    if not os.path.isfile(output):
        return True

    output_mtime = os.path.getmtime(output)

    dep_mtime = 0.0
    for file in dependencies:
        mtime = os.path.getmtime(file)
        if mtime > dep_mtime:
            dep_mtime = mtime

    if dep_mtime > output_mtime:
        return True

    return False


def get_abs_paths(paths: Sequence[PathLike]) -> List[str]:
    """Return list of *paths* in absolute form."""

    paths_abs: List[str] = []
    for path in paths:
        if os.path.isabs(path):
            paths_abs.append(os.path.abspath(path))
        else:
            paths_abs.append(os.path.abspath(os.path.join(os.getcwd(), path)))

    return paths_abs


class Icarus(Simulator):
    @staticmethod
    def simulator_in_path() -> None:
        if shutil.which("iverilog") is None:
            raise SystemExit("ERROR: iverilog exacutable not found!")

    @staticmethod
    def check_toplevel_lang(toplevel_lang: Optional[str]) -> str:
        if toplevel_lang is None or toplevel_lang == "verilog":
            return "verilog"
        else:
            raise ValueError(
                f"iverilog does not support {toplevel_lang!r} as a toplevel_lang"
            )

    @staticmethod
    def get_include_options(includes: Sequence[str]) -> List[str]:
        return ["-I" + dir for dir in includes]

    @staticmethod
    def get_define_options(defines: Sequence[str]) -> List[str]:
        return ["-D" + define for define in defines]

    def get_parameter_options(self, parameters: Mapping[str, object]) -> List[str]:
        assert self.hdl_toplevel is not None
        return [
            f"-P{self.hdl_toplevel}.{name}={value}"
            for name, value in parameters.items()
        ]

    @property
    def sim_file(self) -> PathLike:
        return os.path.join(self.build_dir, "sim.vvp")

    def test_command(self) -> List[Command]:

        return [
            [
                "vvp",
                "-M",
                cocotb.config.libs_dir,
                "-m",
                cocotb.config.lib_name("vpi", "icarus"),
            ]
            + self.sim_args
            + [self.sim_file]
            + self.plus_args
        ]

    def build_command(self) -> List[Command]:

        if self.vhdl_sources:
            raise ValueError("This simulator does not support VHDL")

        cmd = []
        if outdated(self.sim_file, self.verilog_sources) or self.always:

            cmd = [
                ["iverilog", "-o", self.sim_file, "-D", "COCOTB_SIM=1", "-g2012"]
                + self.get_define_options(self.defines)
                + self.get_include_options(self.includes)
                + self.get_parameter_options(self.parameters)
                + self.compile_args
                + self.verilog_sources
            ]

        else:
            print("WARNING: Skipping compilation:" + self.sim_file)

        return cmd


class Questa(Simulator):
    @staticmethod
    def simulator_in_path() -> None:
        if shutil.which("vsim") is None:
            raise SystemExit("ERROR: vsim executable not found!")

    def check_toplevel_lang(self, toplevel_lang: Optional[str]) -> str:
        if toplevel_lang is None:
            if self.vhdl_sources and not self.verilog_sources:
                return "vhdl"
            elif self.verilog_sources and not self.vhdl_sources:
                return "verilog"
            else:
                raise ValueError("Must specify a toplevel_lang in a mixed design")
        elif toplevel_lang in ("verilog", "vhdl"):
            return toplevel_lang
        else:
            raise ValueError(
                f"Questa does not support {toplevel_lang!r} as a toplevel_lang"
            )

    @staticmethod
    def get_include_options(includes: Sequence[str]) -> List[str]:
        return ["+incdir+" + as_tcl_value(dir) for dir in includes]

    @staticmethod
    def get_define_options(defines: Sequence[str]) -> List[str]:
        return ["+define+" + as_tcl_value(define) for define in defines]

    @staticmethod
    def get_parameter_options(parameters: Mapping[str, object]) -> List[str]:
        return ["-g" + name + "=" + str(value) for name, value in parameters.items()]

    def build_command(self) -> List[Command]:

        cmd = []

        if self.vhdl_sources:
            cmd.append(["vlib", as_tcl_value(self.library_name)])
            cmd.append(
                ["vcom", "-mixedsvvh"]
                + ["-work", as_tcl_value(self.library_name)]
                + [as_tcl_value(v) for v in self.compile_args]
                + [as_tcl_value(v) for v in self.vhdl_sources]
            )

        if self.verilog_sources:
            cmd.append(["vlib", as_tcl_value(self.library_name)])
            cmd.append(
                ["vlog", "-mixedsvvh"]
                + ([] if self.always else ["-incr"])
                + ["-work", as_tcl_value(self.library_name)]
                + ["+define+COCOTB_SIM"]
                + ["-sv"]
                + self.get_define_options(self.defines)
                + self.get_include_options(self.includes)
                + [as_tcl_value(v) for v in self.compile_args]
                + [as_tcl_value(v) for v in self.verilog_sources]
            )

        return cmd

    def test_command(self) -> List[Command]:

        cmd = []

        do_script = ""
        if self.waves:
            do_script += "log -recursive /*;"

        if not self.gui:
            do_script += "run -all; quit"

        fli_lib_path = cocotb.config.lib_name_path("fli", "questa")

        if self.toplevel_lang == "vhdl":

            if not os.path.isfile(fli_lib_path):
                raise SystemExit(
                    "ERROR: cocotb was not installed with an FLI library, as the mti.h header could not be located.\n\
                    If you installed an FLI-capable simulator after cocotb, you will need to reinstall cocotb.\n\
                    Please check the cocotb documentation on ModelSim support."
                )

            cmd.append(
                ["vsim"]
                + ["-gui" if self.gui else "-c"]
                + ["-onfinish", "stop" if self.gui else "exit"]
                + [
                    "-foreign",
                    "cocotb_init "
                    + as_tcl_value(cocotb.config.lib_name_path("fli", "questa")),
                ]
                + [as_tcl_value(v) for v in self.sim_args]
                + [as_tcl_value(v) for v in self.get_parameter_options(self.parameters)]
                + [as_tcl_value(self.sim_toplevel)]
                + ["-do", do_script]
            )

            self.env["GPI_EXTRA"] = (
                cocotb.config.lib_name_path("vpi", "questa") + ":cocotbvpi_entry_point"
            )

        else:
            cmd.append(
                ["vsim"]
                + ["-gui" if self.gui else "-c"]
                + ["-onfinish", "stop" if self.gui else "exit"]
                + ["-pli", as_tcl_value(cocotb.config.lib_name_path("vpi", "questa"))]
                + [as_tcl_value(v) for v in self.sim_args]
                + [as_tcl_value(v) for v in self.get_parameter_options(self.parameters)]
                + [as_tcl_value(self.sim_toplevel)]
                + [as_tcl_value(v) for v in self.plus_args]
                + ["-do", do_script]
            )

            if os.path.isfile(fli_lib_path):
                self.env["GPI_EXTRA"] = (
                    cocotb.config.lib_name_path("fli", "questa")
                    + ":cocotbfli_entry_point"
                )
            else:
                print(
                    "WARNING: FLI library not found. Mixed-mode simulation will not be available."
                )

        return cmd


class Ghdl(Simulator):
    @staticmethod
    def simulator_in_path() -> None:
        if shutil.which("ghdl") is None:
            raise SystemExit("ERROR: ghdl executable not found!")

    def check_toplevel_lang(self, toplevel_lang: Optional[str]) -> str:
        if toplevel_lang is None or toplevel_lang == "vhdl":
            return "vhdl"
        else:
            raise ValueError(
                f"GHDL does not support {toplevel_lang!r} as a toplevel_lang"
            )

    @staticmethod
    def get_include_options(includes: Sequence[str]) -> List[str]:
        return [f"-I{dir}" for dir in includes]

    @staticmethod
    def get_define_options(defines: Sequence[str]) -> List[str]:
        return [f"-D{define}" for define in defines]

    @staticmethod
    def get_parameter_options(parameters: Mapping[str, object]) -> List[str]:
        return ["-g" + name + "=" + str(value) for name, value in parameters.items()]

    def build_command(self) -> List[Command]:

        if self.verilog_sources:
            raise ValueError("This simulator does not support Verilog")

        if self.hdl_toplevel is None:
            raise ValueError(
                "This simulator requires the hdl_toplevel parameter to be specified"
            )

        cmd = [
            ["ghdl", "-i"]
            + [f"--work={self.library_name}"]
            + self.compile_args
            + [source_file]
            for source_file in self.vhdl_sources
        ]

        cmd += [
            ["ghdl", "-m"]
            + [f"--work={self.library_name}"]
            + self.compile_args
            + [self.hdl_toplevel]
        ]

        return cmd

    def test_command(self) -> List[Command]:

        cmd = [
            ["ghdl", "-r"]
            + [self.sim_toplevel]
            + ["--vpi=" + cocotb.config.lib_name_path("vpi", "ghdl")]
            + self.sim_args
            + self.get_parameter_options(self.parameters)
        ]

        return cmd


class Riviera(Simulator):
    @staticmethod
    def simulator_in_path() -> None:
        if shutil.which("vsimsa") is None:
            raise SystemExit("ERROR: vsimsa executable not found!")

    def check_toplevel_lang(self, toplevel_lang: Optional[str]) -> str:
        if toplevel_lang is None:
            if self.vhdl_sources and not self.verilog_sources:
                return "vhdl"
            elif self.verilog_sources and not self.vhdl_sources:
                return "verilog"
            else:
                raise ValueError(
                    "Must specify a toplevel_lang in a mixed-language design"
                )
        elif toplevel_lang in ("verilog", "vhdl"):
            return toplevel_lang
        else:
            raise ValueError(
                f"Riviera does not support {toplevel_lang!r} as a toplevel_lang"
            )

    @staticmethod
    def get_include_options(includes: Sequence[str]) -> List[str]:
        return ["+incdir+" + as_tcl_value(dir) for dir in includes]

    @staticmethod
    def get_define_options(defines: Sequence[str]) -> List[str]:
        return ["+define+" + as_tcl_value(define) for define in defines]

    @staticmethod
    def get_parameter_options(parameters: Mapping[str, object]) -> List[str]:
        return ["-g" + name + "=" + str(value) for name, value in parameters.items()]

    def build_command(self) -> List[Command]:

        do_script = "\nonerror {\n quit -code 1 \n} \n"

        out_file = os.path.join(
            self.build_dir, self.library_name, self.library_name + ".lib"
        )

        if outdated(out_file, self.verilog_sources + self.vhdl_sources) or self.always:

            do_script += "alib {RTL_LIBRARY} \n".format(
                RTL_LIBRARY=as_tcl_value(self.library_name)
            )

            if self.vhdl_sources:
                do_script += (
                    "acom -work {RTL_LIBRARY} {EXTRA_ARGS} {VHDL_SOURCES}\n".format(
                        RTL_LIBRARY=as_tcl_value(self.library_name),
                        VHDL_SOURCES=" ".join(
                            as_tcl_value(v) for v in self.vhdl_sources
                        ),
                        EXTRA_ARGS=" ".join(as_tcl_value(v) for v in self.compile_args),
                    )
                )

            if self.verilog_sources:
                do_script += "alog -work {RTL_LIBRARY} +define+COCOTB_SIM -sv {DEFINES} {INCDIR} {EXTRA_ARGS} {VERILOG_SOURCES} \n".format(
                    RTL_LIBRARY=as_tcl_value(self.library_name),
                    VERILOG_SOURCES=" ".join(
                        as_tcl_value(v) for v in self.verilog_sources
                    ),
                    DEFINES=" ".join(self.get_define_options(self.defines)),
                    INCDIR=" ".join(self.get_include_options(self.includes)),
                    EXTRA_ARGS=" ".join(as_tcl_value(v) for v in self.compile_args),
                )
        else:
            print("WARNING: Skipping compilation:" + out_file)

        do_file = tempfile.NamedTemporaryFile(delete=False)
        do_file.write(do_script.encode())
        do_file.close()

        return [["vsimsa"] + ["-do"] + ["do"] + [do_file.name]]

    def test_command(self) -> List[Command]:

        do_script = "\nonerror {\n quit -code 1 \n} \n"

        if self.toplevel_lang == "vhdl":
            do_script += "asim +access +w -interceptcoutput -O2 -loadvhpi {EXT_NAME} {EXTRA_ARGS} {TOPLEVEL} \n".format(
                TOPLEVEL=as_tcl_value(self.sim_toplevel),
                EXT_NAME=as_tcl_value(cocotb.config.lib_name_path("vhpi", "riviera")),
                EXTRA_ARGS=" ".join(
                    as_tcl_value(v)
                    for v in (
                        self.sim_args + self.get_parameter_options(self.parameters)
                    )
                ),
            )

            self.env["GPI_EXTRA"] = (
                cocotb.config.lib_name_path("vpi", "riviera") + ":cocotbvpi_entry_point"
            )
        else:
            do_script += "asim +access +w -interceptcoutput -O2 -pli {EXT_NAME} {EXTRA_ARGS} {TOPLEVEL} {PLUS_ARGS} \n".format(
                TOPLEVEL=as_tcl_value(self.sim_toplevel),
                EXT_NAME=as_tcl_value(cocotb.config.lib_name_path("vpi", "riviera")),
                EXTRA_ARGS=" ".join(
                    as_tcl_value(v)
                    for v in (
                        self.sim_args + self.get_parameter_options(self.parameters)
                    )
                ),
                PLUS_ARGS=" ".join(as_tcl_value(v) for v in self.plus_args),
            )

            self.env["GPI_EXTRA"] = (
                cocotb.config.lib_name_path("vhpi", "riviera")
                + ":cocotbvhpi_entry_point"
            )

        if self.waves:
            do_script += "log -recursive /*;"

        do_script += "run -all \nexit"

        do_file = tempfile.NamedTemporaryFile(delete=False)
        do_file.write(do_script.encode())
        do_file.close()

        return [["vsimsa"] + ["-do"] + ["do"] + [do_file.name]]


class Verilator(Simulator):
    def simulator_in_path(self) -> None:
        executable = shutil.which("verilator")
        if executable is None:
            raise SystemExit("ERROR: verilator executable not found!")
        self.executable = executable

    @staticmethod
    def check_toplevel_lang(toplevel_lang: Optional[str]) -> str:
        if toplevel_lang is None or toplevel_lang == "verilog":
            return "verilog"
        else:
            raise ValueError(
                f"Verilator does not support {toplevel_lang!r} as a toplevel_lang"
            )

    @staticmethod
    def get_include_options(includes: Sequence[str]) -> List[str]:
        return ["-I" + dir for dir in includes]

    @staticmethod
    def get_define_options(defines: Sequence[str]) -> List[str]:
        return ["-D" + define for define in defines]

    @staticmethod
    def get_parameter_options(parameters: Mapping[str, object]) -> List[str]:
        return ["-G" + name + "=" + str(value) for name, value in parameters.items()]

    def build_command(self) -> List[Command]:

        if self.vhdl_sources:
            raise ValueError("This simulator does not support VHDL")

        if self.hdl_toplevel is None:
            raise ValueError(
                "This simulator requires hdl_toplevel parameter to be specified"
            )

        cmd = []

        verilator_cpp = os.path.join(
            os.path.dirname(cocotb.__file__),
            "share",
            "lib",
            "verilator",
            "verilator.cpp",
        )

        cmd.append(
            [
                "perl",
                self.executable,
                "-cc",
                "--exe",
                "-Mdir",
                self.build_dir,
                "-DCOCOTB_SIM=1",
                "--top-module",
                self.hdl_toplevel,
                "--vpi",
                "--public-flat-rw",
                "--prefix",
                "Vtop",
                "-o",
                self.hdl_toplevel,
                "-LDFLAGS",
                "-Wl,-rpath,{LIB_DIR} -L{LIB_DIR} -lcocotbvpi_verilator".format(
                    LIB_DIR=cocotb.config.libs_dir
                ),
            ]
            + self.compile_args
            + self.get_define_options(self.defines)
            + self.get_include_options(self.includes)
            + self.get_parameter_options(self.parameters)
            + [verilator_cpp]
            + self.verilog_sources
        )

        cmd.append(["make", "-C", self.build_dir, "-f", "Vtop.mk"])

        return cmd

    def test_command(self) -> List[Command]:
        out_file = os.path.join(self.build_dir, self.sim_toplevel)
        return [[out_file] + self.plus_args]


class Xcelium(Simulator):
    @staticmethod
    def simulator_in_path() -> None:
        if shutil.which("xrun") is None:
            raise SystemExit("ERROR: xrun executable not found!")

    def check_toplevel_lang(self, toplevel_lang: Optional[str]) -> str:
        if toplevel_lang is None:
            if self.vhdl_sources and not self.verilog_sources:
                return "vhdl"
            elif self.verilog_sources and not self.vhdl_sources:
                return "verilog"
            else:
                raise ValueError("Must specify a toplevel_lang in a mixed design")
        elif toplevel_lang in ("verilog", "vhdl"):
            return toplevel_lang
        else:
            raise ValueError(
                f"Xcelium does not support {toplevel_lang!r} as a toplevel_lang"
            )

    @staticmethod
    def get_include_options(includes: Sequence[str]) -> List[str]:
        return [f"-incdir {include}" for include in includes]

    @staticmethod
    def get_define_options(defines: Sequence[str]) -> List[str]:
        return [f"-define {define}" for define in defines]

    @staticmethod
    def get_parameter_options(parameters: Mapping[str, object]) -> List[str]:
        return [f'-gpg "{name} => {value}"' for name, value in parameters.items()]

    def build_command(self) -> List[Command]:

        self.env["CDS_AUTO_64BIT"] = "all"
        cmd = [
            ["xrun"]
            + ["-logfile xrun_build.log"]
            + ["-elaborate"]
            + [f"-xmlibdirname {self.build_dir}/xrun_snapshot"]
            + ["-licqueue"]
            # TODO: way to switch to these verbose messages?:
            + ["-messages"]
            + ["-gverbose"]  # print assigned generics/parameters
            + ["-plinowarn"]
            # + ["-pliverbose"]
            # + ["-plidebug"]  # Enhance the profile output with PLI info
            # + ["-plierr_verbose"]  # Expand handle info in PLI/VPI/VHPI messages
            # + ["-vpicompat 1800v2005"]  # <1364v1995|1364v2001|1364v2005|1800v2005> Specify the IEEE VPI
            + ["-access +rwc"]
            + [
                "-loadvpi "
                + cocotb.config.lib_name_path("vpi", "xcelium")
                + ":vlog_startup_routines_bootstrap"
            ]
            + [f"-work {self.library_name}"]
            + self.compile_args
            + ["-define COCOTB_SIM"]
            + self.get_define_options(self.defines)
            + self.get_include_options(self.includes)
            + self.get_parameter_options(self.parameters)
            + [f"-top {self.hdl_toplevel}"]
            + self.vhdl_sources
            + self.verilog_sources
        ]

        return cmd

    def test_command(self) -> List[Command]:
        self.env["CDS_AUTO_64BIT"] = "all"

        tmpdir = f"implicit_tmpdir_{self.current_test_name}"
        cmd = [["mkdir", "-p", tmpdir]]

        cmd += [
            ["xrun"]
            + [f"-logfile xrun_{self.current_test_name}.log"]
            + [
                f"-xmlibdirname {self.build_dir}/xrun_snapshot -cds_implicit_tmpdir {tmpdir}"
            ]
            + ["-licqueue"]
            # TODO: way to switch to these verbose messages?:
            + ["-messages"]
            + ["-plinowarn"]
            # + ["-pliverbose"]
            # + ["-plidebug"]  # Enhance the profile output with PLI info
            # + ["-plierr_verbose"]  # Expand handle info in PLI/VPI/VHPI messages
            # + ["-vpicompat 1800v2005"]  # <1364v1995|1364v2001|1364v2005|1800v2005> Specify the IEEE VPI
            + ["-R"]
            + self.sim_args
            + self.plus_args
            + ["-gui" if self.gui else ""]
            + [
                '-input "@probe -create {self.sim_toplevel} -all -depth all"'
                if self.waves
                else ""
            ]
        ]
        self.env["GPI_EXTRA"] = (
            cocotb.config.lib_name_path("vhpi", "xcelium") + ":cocotbvhpi_entry_point"
        )

        return cmd


def get_runner(simulator_name: str) -> Type[Simulator]:

    sim_name = simulator_name.lower()
    supported_sims: Dict[str, Type[Simulator]] = {
        "icarus": Icarus,
        "questa": Questa,
        "ghdl": Ghdl,
        "riviera": Riviera,
        "verilator": Verilator,
        "xcelium": Xcelium,
        # TODO: "vcs": Vcs,
    }
    try:
        return supported_sims[sim_name]
    except KeyError:
        raise NotImplementedError(
            "Set SIM variable. Supported: " + ", ".join(supported_sims)
        ) from None


def clean(recursive: bool = False) -> None:
    dir = os.getcwd()

    def rm_clean() -> None:
        build_dir = os.path.join(dir, "sim_build")
        if os.path.isdir(build_dir):
            print("INFO: Removing:", build_dir)
            shutil.rmtree(build_dir, ignore_errors=True)

    rm_clean()

    if recursive:
        for dir, _, _ in os.walk(dir):
            rm_clean()



# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Potential Ventures Ltd,
#       SolarFlare Communications Inc nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Element, SubElement


class XUnitReporter:
    def __init__(self, filename="results.xml"):
        self.results = Element("testsuites", name="results")
        self.filename = filename

    def add_testsuite(self, **kwargs):
        self.last_testsuite = SubElement(self.results, "testsuite", **kwargs)
        return self.last_testsuite

    def add_testcase(self, testsuite=None, **kwargs):
        if testsuite is None:
            testsuite = self.last_testsuite
        self.last_testcase = SubElement(testsuite, "testcase", **kwargs)
        return self.last_testcase

    def add_property(self, testsuite=None, **kwargs):
        if testsuite is None:
            testsuite = self.last_testsuite
        self.last_property = SubElement(testsuite, "property", **kwargs)
        return self.last_property

    def add_failure(self, testcase=None, **kwargs):
        if testcase is None:
            testcase = self.last_testcase
        SubElement(testcase, "failure", **kwargs)

    def add_skipped(self, testcase=None, **kwargs):
        if testcase is None:
            testcase = self.last_testcase
        SubElement(testcase, "skipped", **kwargs)

    def indent(self, elem, level=0):
        i = "\n" + level * "  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + "  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for elem in elem:
                self.indent(elem, level + 1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i

    def write(self):
        self.indent(self.results)
        ET.ElementTree(self.results).write(self.filename, encoding="UTF-8")



#!/usr/bin/env python

# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Potential Ventures Ltd,
#       SolarFlare Communications Inc nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import random
import re
import warnings

_RESOLVE_TO_0 = "-lL"
_RESOLVE_TO_1 = "hH"
_RESOLVE_TO_CHOICE = "xXzZuUwW"
resolve_x_to = os.getenv("COCOTB_RESOLVE_X", "VALUE_ERROR")


class _ResolveTable(dict):
    """Translation table class for resolving binary strings.

    For use with :func:`str.translate()`, which indexes into table with Unicode ordinals.
    """

    def __init__(self):
        self.update({ord("0"): ord("0"), ord("1"): ord("1")})
        self.update({ord(k): ord("0") for k in _RESOLVE_TO_0})
        self.update({ord(k): ord("1") for k in _RESOLVE_TO_1})

        # Do not resolve if resolve_x_to is not set to one of the supported values
        def no_resolve(key):
            return key

        self.resolve_x = no_resolve

        if resolve_x_to == "VALUE_ERROR":

            def resolve_error(key):
                raise ValueError(
                    "Unresolvable bit in binary string: '{}'".format(chr(key))
                )

            self.resolve_x = resolve_error
        elif resolve_x_to == "ZEROS":
            self.update({ord(k): ord("0") for k in _RESOLVE_TO_CHOICE})
        elif resolve_x_to == "ONES":
            self.update({ord(k): ord("1") for k in _RESOLVE_TO_CHOICE})
        elif resolve_x_to == "RANDOM":

            def resolve_random(key):
                # convert to correct Unicode ordinal:
                # ord('0') = 48
                # ord('1') = 49
                return random.getrandbits(1) + 48

            self.resolve_x = resolve_random

        self._resolve_to_choice = {ord(c) for c in _RESOLVE_TO_CHOICE}

    def __missing__(self, key):
        if key in self._resolve_to_choice:
            return self.resolve_x(key)
        else:
            return key


_resolve_table = _ResolveTable()


def resolve(string):
    return string.translate(_resolve_table)


def _clog2(val):
    if val < 0:
        raise ValueError("_clog2 can't take a negative")
    exp = 0
    while True:
        if (1 << exp) >= val:
            return exp
        exp += 1


class BinaryRepresentation:  # noqa
    UNSIGNED = 0  #: Unsigned format
    SIGNED_MAGNITUDE = 1  #: Sign and magnitude format
    TWOS_COMPLEMENT = 2  #: Two's complement format


class BinaryValue:
    """Representation of values in binary format.

    The underlying value can be set or accessed using these aliasing attributes:

        - :attr:`BinaryValue.integer` is an integer
        - :attr:`BinaryValue.signed_integer` is a signed integer
        - :attr:`BinaryValue.binstr` is a string of ``01xXzZ``
        - :attr:`BinaryValue.buff` is a binary buffer of bytes
        - :attr:`BinaryValue.value` is an integer **deprecated**

    For example:

    >>> vec = BinaryValue()
    >>> vec.integer = 42
    >>> print(vec.binstr)
    101010
    >>> print(vec.buff)
    b'*'

    """

    _permitted_chars = _RESOLVE_TO_0 + _RESOLVE_TO_1 + _RESOLVE_TO_CHOICE + "01"  # noqa

    def __init__(
        self,
        value=None,
        n_bits=None,
        bigEndian=True,
        binaryRepresentation=BinaryRepresentation.UNSIGNED,
        bits=None,
    ):
        """
        Args:
            value (str or int or long, optional): Value to assign to the bus.
            n_bits (int, optional): Number of bits to use for the underlying
                binary representation.
            bigEndian (bool, optional): Interpret the binary as big-endian
                when converting to/from a string buffer.
            binaryRepresentation (BinaryRepresentation): The representation
                of the binary value
                (one of :any:`UNSIGNED`, :any:`SIGNED_MAGNITUDE`, :any:`TWOS_COMPLEMENT`).
                Defaults to unsigned representation.
            bits (int, optional): Deprecated: Compatibility wrapper for :attr:`n_bits`.
        """
        self._str = ""
        self.big_endian = bigEndian
        self.binaryRepresentation = binaryRepresentation

        # bits is the deprecated name for n_bits, allow its use for
        # backward-compat reasons.
        if bits is not None and n_bits is not None:
            raise TypeError("You cannot use n_bits and bits at the same time.")
        if bits is not None:
            warnings.warn(
                "The bits argument to BinaryValue has been renamed to n_bits",
                DeprecationWarning,
                stacklevel=2,
            )
            n_bits = bits

        self._n_bits = n_bits

        self._convert_to = self._convert_to_map[self.binaryRepresentation].__get__(
            self, self.__class__
        )

        self._convert_from = self._convert_from_map[self.binaryRepresentation].__get__(
            self, self.__class__
        )

        if value is not None:
            self.assign(value)

    def assign(self, value):
        """Decides how best to assign the value to the vector.

        Picks from the type of its argument whether to set :attr:`integer`,
        :attr:`binstr`, or :attr:`buff`.

        Args:
            value (str or int or bytes): The value to assign.

        .. versionchanged:: 1.4

            This no longer falls back to setting :attr:`buff` if a :class:`str`
            containing any characters that aren't ``0``, ``1``, ``X`` or ``Z``
            is used, since :attr:`buff` now accepts only :class:`bytes`. Instead,
            an error is raised.
        """
        if isinstance(value, int):
            self.integer = value
        elif isinstance(value, str):
            self.binstr = value
        elif isinstance(value, bytes):
            self.buff = value
        else:
            raise TypeError(
                "value must be int, str, or bytes, not {!r}".format(
                    type(value).__qualname__
                )
            )

    def _convert_to_unsigned(self, x):
        if x == 0:
            return self._adjust_unsigned("")
        x = bin(x)
        if x[0] == "-":
            raise ValueError(
                "Attempt to assigned negative number to unsigned " "BinaryValue"
            )
        return self._adjust_unsigned(x[2:])

    def _convert_to_signed_mag(self, x):
        if x == 0:
            return self._adjust_unsigned("")
        x = bin(x)
        if x[0] == "-":
            binstr = self._adjust_signed_mag("1" + x[3:])
        else:
            binstr = self._adjust_signed_mag("0" + x[2:])
        if self.big_endian:
            binstr = binstr[::-1]
        return binstr

    def _convert_to_twos_comp(self, x):
        if x < 0:
            binstr = bin(2 ** (_clog2(abs(x)) + 1) + x)[2:]
            binstr = self._adjust_twos_comp(binstr)
        elif x == 0:
            binstr = self._adjust_twos_comp("")
        else:
            binstr = self._adjust_twos_comp("0" + bin(x)[2:])
        if self.big_endian:
            binstr = binstr[::-1]
        return binstr

    def _convert_from_unsigned(self, x):
        if not len(x):
            return 0
        return int(x.translate(_resolve_table), 2)

    def _convert_from_signed_mag(self, x):
        if not len(x):
            return 0
        rv = int(self._str[1:].translate(_resolve_table), 2)
        if self._str[0] == "1":
            rv = rv * -1
        return rv

    def _convert_from_twos_comp(self, x):
        if not len(x):
            return 0
        if x[0] == "1":
            binstr = x[1:]
            binstr = self._invert(binstr)
            rv = int(binstr, 2) + 1
            rv = rv * -1
        else:
            rv = int(x.translate(_resolve_table), 2)
        return rv

    _convert_to_map = {
        BinaryRepresentation.UNSIGNED: _convert_to_unsigned,
        BinaryRepresentation.SIGNED_MAGNITUDE: _convert_to_signed_mag,
        BinaryRepresentation.TWOS_COMPLEMENT: _convert_to_twos_comp,
    }

    _convert_from_map = {
        BinaryRepresentation.UNSIGNED: _convert_from_unsigned,
        BinaryRepresentation.SIGNED_MAGNITUDE: _convert_from_signed_mag,
        BinaryRepresentation.TWOS_COMPLEMENT: _convert_from_twos_comp,
    }

    _invert_table = str.maketrans({"0": "1", "1": "0"})

    def _invert(self, x):
        return x.translate(self._invert_table)

    def _adjust_unsigned(self, x):
        if self._n_bits is None:
            return x
        l = len(x)
        if l <= self._n_bits:
            if self.big_endian:
                rv = x + "0" * (self._n_bits - l)
            else:
                rv = "0" * (self._n_bits - l) + x
        elif l > self._n_bits:
            if self.big_endian:
                rv = x[l - self._n_bits :]
            else:
                rv = x[: l - self._n_bits]
            warnings.warn(
                "{}-bit value requested, truncating value {!r} ({} bits) to {!r}".format(
                    self._n_bits, x, l, rv
                ),
                category=RuntimeWarning,
                stacklevel=3,
            )
        return rv

    def _adjust_signed_mag(self, x):
        """Pad/truncate the bit string to the correct length."""
        if self._n_bits is None:
            return x
        l = len(x)
        if l < self._n_bits:
            if self.big_endian:
                rv = x[:-1] + "0" * (self._n_bits - 1 - l)
                rv = rv + x[-1]
            else:
                rv = "0" * (self._n_bits - 1 - l) + x[1:]
                rv = x[0] + rv
        elif l > self._n_bits:
            if self.big_endian:
                rv = x[l - self._n_bits :]
            else:
                rv = x[: -(l - self._n_bits)]
            warnings.warn(
                "{}-bit value requested, truncating value {!r} ({} bits) to {!r}".format(
                    self._n_bits, x, l, rv
                ),
                category=RuntimeWarning,
                stacklevel=3,
            )
        else:
            rv = x
        return rv

    def _adjust_twos_comp(self, x):
        if self._n_bits is None:
            return x
        l = len(x)
        if l == 0:
            rv = x
        elif l < self._n_bits:
            if self.big_endian:
                rv = x + x[-1] * (self._n_bits - l)
            else:
                rv = x[0] * (self._n_bits - l) + x
        elif l > self._n_bits:
            if self.big_endian:
                rv = x[l - self._n_bits :]
            else:
                rv = x[: -(l - self._n_bits)]
            warnings.warn(
                "{}-bit value requested, truncating value {!r} ({} bits) to {!r}".format(
                    self._n_bits, x, l, rv
                ),
                category=RuntimeWarning,
                stacklevel=3,
            )
        else:
            rv = x
        return rv

    @property
    def integer(self):
        """The integer representation of the underlying vector."""
        return self._convert_from(self._str)

    @integer.setter
    def integer(self, val):
        self._str = self._convert_to(val)

    @property
    def value(self):
        """Integer access to the value. **deprecated**"""
        return self.integer

    @value.setter
    def value(self, val):
        self.integer = val

    get_value = value.fget
    set_value = value.fset

    @property
    def signed_integer(self):
        """The signed integer representation of the underlying vector."""
        ival = int(self._str.translate(_resolve_table), 2)
        bits = len(self._str)
        signbit = 1 << (bits - 1)
        if (ival & signbit) == 0:
            return ival
        else:
            return -1 * (1 + (int(~ival) & (signbit - 1)))

    @signed_integer.setter
    def signed_integer(self, val):
        self.integer = val

    get_value_signed = signed_integer.fget

    @property
    def is_resolvable(self) -> bool:
        """
        Return whether the value contains only resolvable (i.e. no "unknown") bits.

        By default the values ``X``, ``Z``, ``U`` and ``W`` are considered unresolvable.
        This can be configured with :envvar:`COCOTB_RESOLVE_X`.

        This is similar to the SystemVerilog Assertion ``$isunknown`` system function
        or the VHDL function ``is_x`` (with an inverted meaning).
        """
        return not any(char in self._str for char in _RESOLVE_TO_CHOICE)

    @property
    def buff(self) -> bytes:
        r"""The value as a binary string buffer.

        >>> BinaryValue("01000001" + "00101111").buff == b"\x41\x2F"
        True

        .. versionchanged:: 1.4
            This changed from :class:`str` to :class:`bytes`.
            Note that for older versions used with Python 2 these types were
            indistinguishable.
        """
        bits = self._str.translate(_resolve_table)

        if len(bits) % 8:
            bits = "0" * (8 - len(bits) % 8) + bits

        buff = []
        while bits:
            byte = bits[:8]
            bits = bits[8:]
            val = int(byte, 2)
            if self.big_endian:
                buff += [val]
            else:
                buff = [val] + buff
        return bytes(buff)

    @buff.setter
    def buff(self, val: bytes):
        if not self.big_endian:
            val = reversed(val)
        self._str = "".join([format(char, "08b") for char in val])
        self._adjust()

    def _adjust(self):
        """Pad/truncate the bit string to the correct length."""
        if self._n_bits is None:
            return
        l = len(self._str)
        if l < self._n_bits:
            if self.big_endian:
                self._str = self._str + "0" * (self._n_bits - l)
            else:
                self._str = "0" * (self._n_bits - l) + self._str
        elif l > self._n_bits:
            rv = self._str[l - self._n_bits :]
            warnings.warn(
                "{}-bit value requested, truncating value {!r} ({} bits) to {!r}".format(
                    self._n_bits, self._str, l, rv
                ),
                category=RuntimeWarning,
                stacklevel=3,
            )
            self._str = rv

    get_buff = buff.fget
    set_buff = buff.fset

    @property
    def binstr(self):
        """The binary representation stored as a string of ``0``, ``1``, and possibly ``x``, ``z``, and other states."""
        return self._str

    _non_permitted_regex = re.compile(f"[^{_permitted_chars}]")

    @binstr.setter
    def binstr(self, string):
        match = self._non_permitted_regex.search(string)
        if match:
            raise ValueError(
                "Attempting to assign character %s to a %s"
                % (match.group(), self.__class__.__name__)
            )
        self._str = string
        self._adjust()

    get_binstr = binstr.fget
    set_binstr = binstr.fset

    def _set_trusted_binstr(self, string):
        self._str = string

    @property
    def n_bits(self):
        """The number of bits of the binary value."""
        return self._n_bits

    def hex(self):
        try:
            return hex(self.integer)
        except Exception:
            return hex(int(self.binstr, 2))

    def __le__(self, other):
        self.assign(other)

    def __str__(self):
        return self.binstr

    def __repr__(self):
        return self.__str__()

    def __bool__(self):
        """Provide boolean testing of a :attr:`binstr`.

        >>> val = BinaryValue("0000")
        >>> if val: print("True")
        ... else:   print("False")
        False
        >>> val.integer = 42
        >>> if val: print("True")
        ... else:   print("False")
        True

        """
        for char in self._str:
            if char == "1":
                return True
        return False

    def __eq__(self, other):
        if isinstance(other, BinaryValue):
            other = other.value
        return self.value == other

    def __ne__(self, other):
        if isinstance(other, BinaryValue):
            other = other.value
        return self.value != other

    def __int__(self):
        return self.integer

    def __long__(self):
        return self.integer

    def __add__(self, other):
        return self.integer + int(other)

    def __iadd__(self, other):
        self.integer = self.integer + int(other)
        return self

    def __radd__(self, other):
        return self.integer + other

    def __sub__(self, other):
        return self.integer - int(other)

    def __isub__(self, other):
        self.integer = self.integer - int(other)
        return self

    def __rsub__(self, other):
        return other - self.integer

    def __mul__(self, other):
        return self.integer * int(other)

    def __imul__(self, other):
        self.integer = self.integer * int(other)
        return self

    def __rmul__(self, other):
        return self.integer * other

    def __floordiv__(self, other):
        return self.integer // int(other)

    def __ifloordiv__(self, other):
        self.integer = self.__floordiv__(other)
        return self

    def __rfloordiv__(self, other):
        return other // self.integer

    def __divmod__(self, other):
        return (self.integer // other, self.integer % other)

    def __rdivmod__(self, other):
        return other // self.integer

    def __mod__(self, other):
        return self.integer % int(other)

    def __imod__(self, other):
        self.integer = self.integer % int(other)
        return self

    def __rmod__(self, other):
        return other % self.integer

    def __pow__(self, other, modulo=None):
        return pow(self.integer, other)

    def __ipow__(self, other):
        self.integer = pow(self.integer, other)
        return self

    def __rpow__(self, other):
        return pow(other, self.integer)

    def __lshift__(self, other):
        return int(self) << int(other)

    def __ilshift__(self, other):
        """Preserves X values"""
        self.binstr = self.binstr[other:] + self.binstr[:other]
        return self

    def __rlshift__(self, other):
        return other << self.integer

    def __rshift__(self, other):
        return int(self) >> int(other)

    def __irshift__(self, other):
        """Preserves X values"""
        self.binstr = self.binstr[-other:] + self.binstr[:-other]
        return self

    def __rrshift__(self, other):
        return other >> self.integer

    def __and__(self, other):
        return self.integer & other

    def __iand__(self, other):
        self.integer &= other
        return self

    def __rand__(self, other):
        return self.integer & other

    def __xor__(self, other):
        return self.integer ^ other

    def __ixor__(self, other):
        self.integer ^= other
        return self

    def __rxor__(self, other):
        return self.__xor__(other)

    def __or__(self, other):
        return self.integer | other

    def __ior__(self, other):
        self.integer |= other
        return self

    def __ror__(self, other):
        return self.__or__(other)

    def __div__(self, other):
        return self.integer / other

    def __idiv__(self, other):
        self.integer /= other
        return self

    def __rdiv__(self, other):
        return other / self.integer

    def __neg__(self):
        return -self.integer

    def __pos__(self):
        return +self.integer

    def __abs__(self):
        return abs(self.integer)

    def __invert__(self):
        """Preserves X values"""
        return self._invert(self.binstr)

    def __oct__(self):
        return oct(self.integer)

    def __hex__(self):
        return hex(self.integer)

    def __index__(self):
        return self.integer

    def __len__(self):
        return len(self.binstr)

    def __getitem__(self, key):
        """BinaryValue uses Verilog/VHDL style slices as opposed to Python
        style"""
        if isinstance(key, slice):
            first, second = key.start, key.stop
            if self.big_endian:
                if first < 0 or second < 0:
                    raise IndexError("BinaryValue does not support negative " "indices")
                if second > self._n_bits - 1:
                    raise IndexError("High index greater than number of bits.")
                if first > second:
                    raise IndexError(
                        "Big Endian indices must be specified " "low to high"
                    )
                _binstr = self.binstr[first : (second + 1)]
            else:
                if first < 0 or second < 0:
                    raise IndexError("BinaryValue does not support negative " "indices")
                if first > self._n_bits - 1:
                    raise IndexError("High index greater than number of bits.")
                if second > first:
                    raise IndexError(
                        "Litte Endian indices must be specified " "high to low"
                    )
                high = self._n_bits - second
                low = self._n_bits - 1 - first
                _binstr = self.binstr[low:high]
        else:
            index = key
            if index > self._n_bits - 1:
                raise IndexError("Index greater than number of bits.")
            if self.big_endian:
                _binstr = self.binstr[index]
            else:
                _binstr = self.binstr[self._n_bits - 1 - index]
        rv = BinaryValue(
            n_bits=len(_binstr),
            bigEndian=self.big_endian,
            binaryRepresentation=self.binaryRepresentation,
        )
        rv.binstr = _binstr
        return rv

    def __setitem__(self, key, val):
        """BinaryValue uses Verilog/VHDL style slices as opposed to Python style."""
        if not isinstance(val, str) and not isinstance(val, int):
            raise TypeError("BinaryValue slices only accept string or integer values")

        # convert integer to string
        if isinstance(val, int):
            if isinstance(key, slice):
                num_slice_bits = abs(key.start - key.stop) + 1
            else:
                num_slice_bits = 1
            if val < 0:
                raise ValueError("Integer must be positive")
            if val >= 2**num_slice_bits:
                raise ValueError(
                    "Integer is too large for the specified slice " "length"
                )
            val = "{:0{width}b}".format(val, width=num_slice_bits)

        if isinstance(key, slice):
            first, second = key.start, key.stop

            if self.big_endian:
                if first < 0 or second < 0:
                    raise IndexError("BinaryValue does not support negative " "indices")
                if second > self._n_bits - 1:
                    raise IndexError("High index greater than number of bits.")
                if first > second:
                    raise IndexError(
                        "Big Endian indices must be specified " "low to high"
                    )
                if len(val) > (second + 1 - first):
                    raise ValueError("String length must be equal to slice " "length")
                slice_1 = self.binstr[:first]
                slice_2 = self.binstr[second + 1 :]
                self.binstr = slice_1 + val + slice_2
            else:
                if first < 0 or second < 0:
                    raise IndexError("BinaryValue does not support negative " "indices")
                if first > self._n_bits - 1:
                    raise IndexError("High index greater than number of bits.")
                if second > first:
                    raise IndexError(
                        "Litte Endian indices must be specified " "high to low"
                    )
                high = self._n_bits - second
                low = self._n_bits - 1 - first
                if len(val) > (high - low):
                    raise ValueError("String length must be equal to slice " "length")
                slice_1 = self.binstr[:low]
                slice_2 = self.binstr[high:]
                self.binstr = slice_1 + val + slice_2
        else:
            if len(val) != 1:
                raise ValueError("String length must be equal to slice " "length")
            index = key
            if index > self._n_bits - 1:
                raise IndexError("Index greater than number of bits.")
            if self.big_endian:
                self.binstr = self.binstr[:index] + val + self.binstr[index + 1 :]
            else:
                self.binstr = (
                    self.binstr[0 : self._n_bits - index - 1]
                    + val
                    + self.binstr[self._n_bits - index : self._n_bits]
                )


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Potential Ventures Ltd,
#       SolarFlare Communications Inc nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys

# TODO: Could use cStringIO?
import traceback
import warnings
from io import StringIO

"""Exceptions and functions for simulation result handling."""


def raise_error(obj, msg):
    """Create a :exc:`TestError` exception and raise it after printing a traceback.

    .. deprecated:: 1.3
        Raise a standard Python exception instead of calling this function.
        A stacktrace will be printed by cocotb automatically if the exception is unhandled.

    Args:
        obj: Object with a log method.
        msg (str): The log message.
    """
    warnings.warn(
        "``raise_error`` is deprecated - raise a standard Exception instead",
        DeprecationWarning,
        stacklevel=2,
    )
    _raise_error(obj, msg)


def _raise_error(obj, msg):
    exc_info = sys.exc_info()
    buff = StringIO()
    traceback.print_exception(*exc_info, file=buff)
    obj.log.error(f"{msg}\n{buff.getvalue()}")
    exception = TestError(msg)
    exception.stderr.write(buff.getvalue())
    raise exception


def create_error(obj, msg):
    """Like :func:`raise_error`, but return the exception rather than raise it,
    simply to avoid too many levels of nested `try/except` blocks.

    .. deprecated:: 1.3
        Raise a standard Python exception instead of calling this function.

    Args:
        obj: Object with a log method.
        msg (str): The log message.
    """
    warnings.warn(
        "``create_error`` is deprecated - raise a standard Exception instead",
        DeprecationWarning,
        stacklevel=2,
    )
    try:
        # use the private version to avoid multiple warnings
        _raise_error(obj, msg)
    except TestError as error:
        return error
    return TestError("Creating error traceback failed")


class ReturnValue(Exception):
    """
    Helper exception needed for Python versions prior to 3.3.

    .. deprecated:: 1.4
        Use a :keyword:`return` statement instead; this works in all supported versions of Python.
    """

    def __init__(self, retval):
        warnings.warn(
            "``ReturnValue`` is deprecated, use a normal return statement instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.retval = retval


class TestComplete(Exception):
    """
    Exception showing that the test was completed. Sub-exceptions detail the exit status.

    .. deprecated:: 1.6.0
        The ``stdout`` and ``stderr`` attributes.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__stdout = StringIO()
        self.__stderr = StringIO()

    @staticmethod
    def __deprecated(which: str) -> None:
        warnings.warn(
            f"Attribute {which} is deprecated and will be removed in the next major release",
            DeprecationWarning,
            stacklevel=3,
        )

    @property
    def stdout(self) -> StringIO:
        self.__deprecated("stdout")
        return self.__stdout

    @stdout.setter
    def stdout(self, new_value: StringIO) -> None:
        self.__deprecated("stdout")
        self.__stdout = new_value

    @property
    def stderr(self) -> StringIO:
        self.__deprecated("stderr")
        return self.__stderr

    @stderr.setter
    def stderr(self, new_value: StringIO) -> None:
        self.__deprecated("stderr")
        self.__stderr = new_value


class ExternalException(Exception):
    """Exception thrown by :class:`cocotb.external` functions."""

    def __init__(self, exception):
        self.exception = exception


class TestError(TestComplete):
    """
    Exception showing that the test was completed with severity Error.

    .. deprecated:: 1.5
        Raise a standard Python exception instead.
        A stacktrace will be printed by cocotb automatically if the exception is unhandled.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "TestError is deprecated - raise a standard Exception instead",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)


class TestFailure(TestComplete, AssertionError):
    """
    Exception showing that the test was completed with severity Failure.

    .. deprecated:: 1.6.0
        Use a standard ``assert`` statement instead of raising this exception.
        Use ``expect_fail`` rather than ``expect_error`` with this exception in the
        :class:`cocotb.test` decorator.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "TestFailure is deprecated, use an ``assert`` statement instead",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)


class TestSuccess(TestComplete):
    """Exception showing that the test was completed successfully."""

    pass


class SimFailure(TestComplete):
    """Exception showing that the simulator exited unsuccessfully."""

    pass


class SimTimeoutError(TimeoutError):
    """Exception for when a timeout, in terms of simulation time, occurs."""

    pass



import importlib
import os
from functools import reduce
from types import ModuleType
from typing import Callable, Tuple


def load_entry() -> Tuple[ModuleType, Callable]:
    """Gather entry point information by parsing :envvar:`PYGPI_ENTRY_POINT`."""
    entry_point_str = os.environ.get(
        "PYGPI_ENTRY_POINT", "cocotb:_initialise_testbench"
    )
    try:
        if ":" not in entry_point_str:
            raise ValueError(
                "Invalid PYGPI_ENTRY_POINT, missing entry function (no colon)."
            )
        entry_module_str, entry_func_str = entry_point_str.split(":", 1)
        entry_module = importlib.import_module(entry_module_str)
        entry_func = reduce(getattr, entry_func_str.split("."), entry_module)
    except Exception as e:
        raise RuntimeError(
            "Failure to parse PYGPI_ENTRY_POINT ('{}')".format(entry_point_str)
        ) from e
    else:
        return entry_module, entry_func



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause

import functools
import warnings
from typing import Any, Callable, Type, TypeVar

AnyCallableT = TypeVar("AnyCallableT", bound=Callable[..., object])


def deprecated(
    msg: str, category: Type[Warning] = DeprecationWarning
) -> Callable[[AnyCallableT], AnyCallableT]:
    """Emits a DeprecationWarning when the decorated function is called.

    This decorator works on normal functions, methods, and properties.
    Usage on properties requires the ``@property`` decorator to appear outside the
    ``@deprecated`` decorator.
    Concrete classes can be deprecated by decorating their ``__init__`` or ``__new__``
    method.

    Args:
        msg: the deprecation message
        category: the warning class to use
    """

    def decorator(f: AnyCallableT) -> AnyCallableT:
        @functools.wraps(f)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            warnings.warn(msg, category=category, stacklevel=2)
            return f(*args, **kwargs)

        return wrapper

    return decorator



# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Potential Ventures Ltd,
#       SolarFlare Communications Inc nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import cherrypy
import dowser


def start(port):
    cherrypy.tree.mount(dowser.Root())
    cherrypy.config.update({"environment": "embedded", "server.socket_port": port})
    cherrypy.engine.start()



# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Potential Ventures Ltd,
#       SolarFlare Communications Inc nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""A clock class."""

import itertools
import warnings

from cocotb.log import SimLog
from cocotb.triggers import Timer
from cocotb.utils import get_sim_steps, get_time_from_sim_steps, lazy_property


class BaseClock:
    """Base class to derive from."""

    def __init__(self, signal):
        self.signal = signal

    @lazy_property
    def log(self):
        return SimLog("cocotb.{}.{}".format(type(self).__qualname__, self.signal._name))


class Clock(BaseClock):
    r"""Simple 50:50 duty cycle clock driver.

    Instances of this class should call its :meth:`start` method
    and pass the coroutine object to one of the functions in :ref:`task-management`.

    This will create a clocking task that drives the signal at the
    desired period/frequency.

    Example:

    .. code-block:: python

        c = Clock(dut.clk, 10, 'ns')
        await cocotb.start(c.start())

    Args:
        signal: The clock pin/signal to be driven.
        period (int): The clock period. Must convert to an even number of
            timesteps.
        units (str, optional): One of
            ``'step'``, ``'fs'``, ``'ps'``, ``'ns'``, ``'us'``, ``'ms'``, ``'sec'``.
            When *units* is ``'step'``,
            the timestep is determined by the simulator (see :make:var:`COCOTB_HDL_TIMEPRECISION`).

    If you need more features like a phase shift and an asymmetric duty cycle,
    it is simple to create your own clock generator (that you then :func:`~cocotb.start`):

    .. code-block:: python

        async def custom_clock():
            # pre-construct triggers for performance
            high_time = Timer(high_delay, units="ns")
            low_time = Timer(low_delay, units="ns")
            await Timer(initial_delay, units="ns")
            while True:
                dut.clk.value = 1
                await high_time
                dut.clk.value = 0
                await low_time

    If you also want to change the timing during simulation,
    use this slightly more inefficient example instead where
    the :class:`Timer`\ s inside the while loop are created with
    current delay values:

    .. code-block:: python

        async def custom_clock():
            while True:
                dut.clk.value = 1
                await Timer(high_delay, units="ns")
                dut.clk.value = 0
                await Timer(low_delay, units="ns")

        high_delay = low_delay = 100
        await cocotb.start(custom_clock())
        await Timer(1000, units="ns")
        high_delay = low_delay = 10  # change the clock speed
        await Timer(1000, units="ns")

    .. versionchanged:: 1.5
        Support ``'step'`` as the *units* argument to mean "simulator time step".

    .. deprecated:: 1.5
        Using ``None`` as the *units* argument is deprecated, use ``'step'`` instead.
    """

    def __init__(self, signal, period, units="step"):
        BaseClock.__init__(self, signal)
        if units is None:
            warnings.warn(
                'Using units=None is deprecated, use units="step" instead.',
                DeprecationWarning,
                stacklevel=2,
            )
            units = "step"  # don't propagate deprecated value
        self.period = get_sim_steps(period, units)
        self.half_period = get_sim_steps(period / 2.0, units)
        self.frequency = 1.0 / get_time_from_sim_steps(self.period, units="us")
        self.hdl = None
        self.signal = signal
        self.coro = None
        self.mcoro = None

    async def start(self, cycles=None, start_high=True):
        r"""Clocking coroutine.  Start driving your clock by :func:`cocotb.start`\ ing a
        call to this.

        Args:
            cycles (int, optional): Cycle the clock *cycles* number of times,
                or if ``None`` then cycle the clock forever.
                Note: ``0`` is not the same as ``None``, as ``0`` will cycle no times.
            start_high (bool, optional): Whether to start the clock with a ``1``
                for the first half of the period.
                Default is ``True``.

                .. versionadded:: 1.3
        """
        t = Timer(self.half_period)
        if cycles is None:
            it = itertools.count()
        else:
            it = range(cycles)

        # branch outside for loop for performance (decision has to be taken only once)
        if start_high:
            for _ in it:
                self.signal.value = 1
                await t
                self.signal.value = 0
                await t
        else:
            for _ in it:
                self.signal.value = 0
                await t
                self.signal.value = 1
                await t

    def __str__(self):
        return type(self).__qualname__ + "(%3.1f MHz)" % self.frequency



# Copyright cocotb contributors
# Licensed under the Revised BSD License, see LICENSE for details.
# SPDX-License-Identifier: BSD-3-Clause
import typing
from itertools import chain

from cocotb.types.range import Range

T = typing.TypeVar("T")
S = typing.TypeVar("S")
Self = typing.TypeVar("Self", bound="Array[typing.Any]")


class Array(typing.Reversible[T], typing.Collection[T]):
    r"""
    Fixed-size, arbitrarily-indexed, homogeneous collection type.

    Arrays are similar to, but different from Python :class:`list`\ s.
    An array can store values of any type or values of multiple types at a time, just like a :class:`list`.
    Unlike :class:`list`\ s, an array's size cannot change.

    The indexes of an array can start or end at any integer value, they are not limited to 0-based indexing.
    Indexing schemes can be either ascending or descending in value.
    An array's indexes are described using a :class:`~cocotb.types.Range` object.
    Initial values are treated as iterables, which are copied into an internal buffer.

    .. code-block:: python3

        >>> Array("1234")  # the 0-based range `(0, len(value)-1)` is inferred
        Array(['1', '2', '3', '4'], Range(0, 'to', 3))

        >>> Array([1, True, None, "example"], Range(-2, 1))  # initial value and range lengths must be equal
        Array([1, True, None, 'example'], Range(-2, 'to', 1))

    Arrays also support "null" ranges; "null" arrays have zero length and cannot be indexed.

    .. code-block:: python3

        >>> Array([], range=Range(1, "to", 0))
        Array([], Range(1, 'to', 0))

    Indexing and slicing is very similar to :class:`list`\ s, but it uses the indexing scheme specified.
    Slicing, just like the :class:`~cocotb.types.Range` object uses an inclusive right bound,
    which is commonly seen in HDLs.
    Like :class:`list`\ s, if a start or stop index is not specified, it is inferred as the start or end of the array.
    Slicing an array returns a new :class:`~cocotb.types.Array` object, whose bounds are the slice indexes.

    .. code-block:: python3

        >>> a = Array("1234abcd")
        >>> a[7]
        'd'
        >>> a[2:5]
        Array(['3', '4', 'a', 'b'], Range(2, 'to', 5))
        >>> a[2:5] = reversed(a[2:5])
        >>> "".join(a)
        '12ba43cd'

        >>> b = Array("1234", Range(0, -3))
        >>> b[-2]
        '3'
        >>> b[-1:]
        Array(['2', '3', '4'], Range(-1, 'downto', -3))
        >>> b[:] = reversed(b)
        >>> b
        Array(['4', '3', '2', '1'], Range(0, 'downto', -3))

    .. warning::
        Arrays behave differently in certain situations than Python's builtin sequence types (:class:`list`, :class:`tuple`, etc.).

        - Arrays are not necessarily 0-based and slices use inclusive right bounds,
          so many functions that work on Python sequences by index (like :mod:`bisect`) may not work on arrays.
        - Slice indexes must be specified in the same direction as the array and do not support specifying a "step".
        - When setting a slice, the new value must be an iterable of the same size as the slice.
        - Negative indexes are *not* treated as an offset from the end of the array, but are treated literally.

    Arrays are equal to other arrays of the same length with the same values (structural equality).
    Bounds do not matter for equality.

    .. code-block:: python3

        >>> a = Array([1, 1, 2, 3, 5], Range(4, "downto", 0))
        >>> b = Array([1, 1, 2, 3, 5], Range(-2, "to", 2))
        >>> a == b
        True

    You can change the bounds of an array by setting the :attr:`range` to a new value.
    The new bounds must be the same length of the array.

    .. code-block:: python3

        >>> a = Array("1234")
        >>> a.range
        Range(0, 'to', 3)
        >>> a.range = Range(3, 'downto', 0)
        >>> a.range
        Range(3, 'downto', 0)

    Arrays support the methods and semantics defined by :class:`collections.abc.Sequence`.

    .. code-block:: python

        >>> a = Array("stuff", Range(2, "downto", -2))
        >>> len(a)
        5
        >>> "t" in a
        True
        >>> a.index("u")
        0
        >>> for c in a:
        ...     print(c)
        s
        t
        u
        f
        f

    Args:
        value: Initial value for the array.
        range: Indexing scheme of the array.

    Raises:
        ValueError: When argument values cannot be used to construct an array.
        TypeError: When invalid argument types are used.
    """

    __slots__ = (
        "_value",
        "_range",
    )

    def __init__(self, value: typing.Iterable[T], range: typing.Optional[Range] = None):
        self._value = list(value)
        if range is None:
            self._range = Range(0, "to", len(self._value) - 1)
        else:
            self._range = range
            if len(self._value) != len(self._range):
                raise ValueError(
                    "init value of length {!r} does not fit in {!r}".format(
                        len(self._value), self._range
                    )
                )

    @property
    def left(self) -> int:
        """Leftmost index of the array."""
        return self.range.left

    @property
    def direction(self) -> str:
        """``"to"`` if indexes are ascending, ``"downto"`` otherwise."""
        return self.range.direction

    @property
    def right(self) -> int:
        """Rightmost index of the array."""
        return self.range.right

    @property
    def range(self) -> Range:
        """:class:`Range` of the indexes of the array."""
        return self._range

    @range.setter
    def range(self, new_range: Range) -> None:
        """Sets a new indexing scheme on the array, must be the same size"""
        if not isinstance(new_range, Range):
            raise TypeError("range argument must be of type 'Range'")
        if len(new_range) != len(self):
            raise ValueError(
                f"{new_range!r} not the same length as old range ({self._range!r})."
            )
        self._range = new_range

    def __len__(self) -> int:
        return len(self.range)

    def __iter__(self) -> typing.Iterator[T]:
        return iter(self._value)

    def __reversed__(self) -> typing.Iterator[T]:
        return reversed(self._value)

    def __contains__(self, item: object) -> bool:
        return item in self._value

    def __eq__(self, other: object) -> bool:
        if isinstance(other, type(self)):
            return self._value == other._value
        return NotImplemented

    @typing.overload
    def __getitem__(self, item: int) -> T:
        ...

    @typing.overload
    def __getitem__(self, item: slice) -> "Array[T]":
        ...

    def __getitem__(
        self, item: typing.Union[int, slice]
    ) -> typing.Union[T, "Array[T]"]:
        if isinstance(item, int):
            idx = self._translate_index(item)
            return self._value[idx]
        elif isinstance(item, slice):
            start = item.start if item.start is not None else self.left
            stop = item.stop if item.stop is not None else self.right
            if item.step is not None:
                raise IndexError("do not specify step")
            start_i = self._translate_index(start)
            stop_i = self._translate_index(stop)
            if start_i > stop_i:
                raise IndexError(
                    "slice [{}:{}] direction does not match array direction [{}:{}]".format(
                        start, stop, self.left, self.right
                    )
                )
            value = self._value[start_i : stop_i + 1]
            range = Range(start, self.direction, stop)
            return type(self)(value=value, range=range)
        raise TypeError(
            "indexes must be ints or slices, not {}".format(type(item).__name__)
        )

    @typing.overload
    def __setitem__(self, item: int, value: T) -> None:
        ...

    @typing.overload
    def __setitem__(self, item: slice, value: typing.Iterable[T]) -> None:
        ...

    def __setitem__(
        self, item: typing.Union[int, slice], value: typing.Union[T, typing.Iterable[T]]
    ) -> None:
        if isinstance(item, int):
            idx = self._translate_index(item)
            self._value[idx] = typing.cast(T, value)
        elif isinstance(item, slice):
            start = item.start if item.start is not None else self.left
            stop = item.stop if item.stop is not None else self.right
            if item.step is not None:
                raise IndexError("do not specify step")
            start_i = self._translate_index(start)
            stop_i = self._translate_index(stop)
            if start_i > stop_i:
                raise IndexError(
                    "slice [{}:{}] direction does not match array direction [{}:{}]".format(
                        start, stop, self.left, self.right
                    )
                )
            value = list(typing.cast(typing.Iterable[T], value))
            if len(value) != (stop_i - start_i + 1):
                raise ValueError(
                    "value of length {!r} will not fit in slice [{}:{}]".format(
                        len(value), start, stop
                    )
                )
            self._value[start_i : stop_i + 1] = value
        else:
            raise TypeError(
                "indexes must be ints or slices, not {}".format(type(item).__name__)
            )

    def __repr__(self) -> str:
        return "{}({!r}, {!r})".format(type(self).__name__, self._value, self._range)

    def __concat__(self: Self, other: Self) -> Self:
        if isinstance(other, type(self)):
            return type(self)(chain(self, other))
        return NotImplemented

    def __rconcat__(self: Self, other: Self) -> Self:
        if isinstance(other, type(self)):
            return type(self)(chain(other, self))
        return NotImplemented

    def index(
        self,
        value: T,
        start: typing.Optional[int] = None,
        stop: typing.Optional[int] = None,
    ) -> int:
        """
        Return index of first occurrence of *value*.

        Raises :exc:`IndexError` if the value is not found.
        Search only within *start* and *stop* if given.
        """
        if start is None:
            start = self.left
        if stop is None:
            stop = self.right
        for i in Range(start, self.direction, stop):
            if self[i] == value:
                return i
        raise IndexError(f"{value!r} not in array")

    def count(self, value: T) -> int:
        """Return number of occurrences of *value*."""
        return self._value.count(value)

    def _translate_index(self, item: int) -> int:
        try:
            return self._range.index(item)
        except ValueError:
            raise IndexError(f"index {item} out of range") from None



#!/usr/bin/env python
###############################################################################
# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Potential Ventures Ltd,
#       SolarFlare Communications Inc nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
###############################################################################

import sys

if sys.version_info[:2] < (3, 6):
    msg = [
        "This version of cocotb requires at least Python 3.6,",
        "you are running Python %d.%d.%d."
        % (sys.version_info[0], sys.version_info[1], sys.version_info[2]),
    ]
    if sys.version_info[0] == 2:
        msg += [
            "If you have Python 3 installed on your machine try ",
            "using 'python3 -m pip' instead of 'pip' to install cocotb.",
        ]
    msg += [
        "For more information please refer to the documentation at ",
        "https://cocotb.readthedocs.io.",
    ]

    raise SystemExit("\n".join(msg))

import logging
from io import StringIO
from os import path, walk

from setuptools import find_packages, setup

# Note: cocotb is not installed properly yet and is missing dependencies and binaries
# We can still import other files next to setup.py, as long as they're in MANIFEST.in
# The below line is necessary for PEP517 support
sys.path.append(path.dirname(__file__))
from cocotb_build_libs import build_ext, get_ext  # noqa: E402


def read_file(fname):
    with open(path.join(path.dirname(__file__), fname), encoding="utf8") as f:
        return f.read()


def package_files(directory):
    paths = []
    for (fpath, directories, filenames) in walk(directory):
        for filename in filenames:
            paths.append(path.join("..", fpath, filename))
    return paths


# this sets the __version__ variable
exec(read_file(path.join("cocotb", "_version.py")))

# store log from build_libs and display at the end in verbose mode
# see https://github.com/pypa/pip/issues/6634
log_stream = StringIO()
handler = logging.StreamHandler(log_stream)
log = logging.getLogger("cocotb._build_libs")
log.setLevel(logging.INFO)
log.addHandler(handler)

setup(
    name="cocotb",
    cmdclass={"build_ext": build_ext},
    version=__version__,  # noqa: F821
    description="cocotb is a coroutine based cosimulation library for writing VHDL and Verilog testbenches in Python.",
    url="https://www.cocotb.org",
    license="BSD",
    long_description=read_file("README.md"),
    long_description_content_type="text/markdown",
    author="Chris Higgs, Stuart Hodgson",
    maintainer="cocotb contributors",
    maintainer_email="cocotb@lists.librecores.org",
    install_requires=[],
    python_requires=">=3.6",
    packages=find_packages(),
    package_data={
        "cocotb": (
            package_files("cocotb/share/makefiles")
            + package_files("cocotb/share/include")  # noqa: W504
            + package_files("cocotb/share/def")  # noqa: W504
            + package_files("cocotb/share/lib/verilator")  # noqa: W504
        )
    },
    ext_modules=get_ext(),
    entry_points={
        "console_scripts": [
            "cocotb-config=cocotb.config:main",
        ]
    },
    platforms="any",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: BSD License",
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
        "Framework :: cocotb",
    ],
    # these appear in the sidebar on PyPI
    project_urls={
        "Bug Tracker": "https://github.com/cocotb/cocotb/issues",
        "Source Code": "https://github.com/cocotb/cocotb",
        "Documentation": "https://docs.cocotb.org",
    },
    extras_require={"bus": ["cocotb_bus"]},
)

print(log_stream.getvalue())



# Copyright (c) cocotb contributors
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Backports and compatibility shims for newer python features.

These are for internal use - users should use a third party library like `six`
if they want to use these shims in their own code
"""
import sys


# backport of Python 3.7's contextlib.nullcontext
class nullcontext:
    """Context manager that does no additional processing.
    Used as a stand-in for a normal context manager, when a particular
    block of code is only sometimes used with a normal context manager:

    cm = optional_cm if condition else nullcontext()
    with cm:
        # Perform operation, using optional_cm if condition is True
    """

    def __init__(self, enter_result=None):
        self.enter_result = enter_result

    def __enter__(self):
        return self.enter_result

    def __exit__(self, *excinfo):
        pass


# On python 3.7 onwards, `dict` is guaranteed to preserve insertion order.
# Since `OrderedDict` is a little slower that `dict`, we prefer the latter
# when possible.
if sys.version_info[:2] >= (3, 7):
    insertion_ordered_dict = dict
else:
    import collections

    insertion_ordered_dict = collections.OrderedDict



# Copyright (c) 2013, 2018 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Potential Ventures Ltd,
#       SolarFlare Communications Inc nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
    Some constants for doing ANSI stuff.
"""
# flake8: noqa (skip this file for flake8: pypi.python.org/pypi/flake8)
_ESCAPE = "\033["

# see https://en.wikipedia.org/wiki/ANSI_escape_code#Colors

DEFAULT_FG = _ESCAPE + "39m"
DEFAULT_BG = _ESCAPE + "49m"
DEFAULT = DEFAULT_BG + DEFAULT_FG

BLACK_FG = _ESCAPE + "30m"
RED_FG = _ESCAPE + "31m"
GREEN_FG = _ESCAPE + "32m"
YELLOW_FG = _ESCAPE + "33m"
BLUE_FG = _ESCAPE + "34m"
MAGENTA_FG = _ESCAPE + "35m"
CYAN_FG = _ESCAPE + "36m"
WHITE_FG = _ESCAPE + "37m"

BLACK_BG = _ESCAPE + "40m"
RED_BG = _ESCAPE + "41m"
GREEN_BG = _ESCAPE + "42m"
YELLOW_BG = _ESCAPE + "43m"
BLUE_BG = _ESCAPE + "44m"
MAGENTA_BG = _ESCAPE + "45m"
CYAN_BG = _ESCAPE + "46m"
WHITE_BG = _ESCAPE + "47m"

BRIGHT_BLACK_FG = _ESCAPE + "90m"
BRIGHT_RED_FG = _ESCAPE + "91m"
BRIGHT_GREEN_FG = _ESCAPE + "92m"
BRIGHT_YELLOW_FG = _ESCAPE + "93m"
BRIGHT_BLUE_FG = _ESCAPE + "94m"
BRIGHT_MAGENTA_FG = _ESCAPE + "95m"
BRIGHT_CYAN_FG = _ESCAPE + "96m"
BRIGHT_WHITE_FG = _ESCAPE + "97m"

BRIGHT_BLACK_BG = _ESCAPE + "100m"
BRIGHT_RED_BG = _ESCAPE + "101m"
BRIGHT_GREEN_BG = _ESCAPE + "102m"
BRIGHT_YELLOW_BG = _ESCAPE + "103m"
BRIGHT_BLUE_BG = _ESCAPE + "104m"
BRIGHT_MAGENTA_BG = _ESCAPE + "105m"
BRIGHT_CYAN_BG = _ESCAPE + "106m"
BRIGHT_WHITE_BG = _ESCAPE + "107m"


COLOR_DEFAULT = DEFAULT

COLOR_WARNING = YELLOW_FG
COLOR_ERROR = RED_FG
COLOR_CRITICAL = RED_BG + BLACK_FG
COLOR_TEST = BLUE_FG
COLOR_PASSED = GREEN_FG
COLOR_SKIPPED = YELLOW_FG
COLOR_FAILED = RED_FG

COLOR_HILITE_SUMMARY = WHITE_FG + RED_BG
COLOR_HILITE_HEXDIFF_DEFAULT = YELLOW_FG
COLOR_HILITE_HEXDIFF_1 = CYAN_FG
COLOR_HILITE_HEXDIFF_2 = RED_FG
COLOR_HILITE_HEXDIFF_3 = MAGENTA_BG
COLOR_HILITE_HEXDIFF_4 = CYAN_BG + BLACK_FG



# Copyright (c) 2013 Potential Ventures Ltd
# Copyright (c) 2013 SolarFlare Communications Inc
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Potential Ventures Ltd,
#       SolarFlare Communications Inc nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL POTENTIAL VENTURES LTD BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import collections.abc
import functools
import inspect
import os
import sys
import typing
import warnings
from asyncio import CancelledError, InvalidStateError

import cocotb
import cocotb.triggers
from cocotb import outcomes
from cocotb.log import SimLog
from cocotb.result import ReturnValue
from cocotb.utils import extract_coro_stack, lazy_property, remove_traceback_frames

T = typing.TypeVar("T")
Self = typing.TypeVar("Self")

# Sadly the Python standard logging module is very slow so it's better not to
# make any calls by testing a boolean flag first
if "COCOTB_SCHEDULER_DEBUG" in os.environ:
    _debug = True
else:
    _debug = False


def public(f):
    """Use a decorator to avoid retyping function/class names.

    * Based on an idea by Duncan Booth:
    http://groups.google.com/group/comp.lang.python/msg/11cbb03e09611b8a
    * Improved via a suggestion by Dave Angel:
    http://groups.google.com/group/comp.lang.python/msg/3d400fb22d8a42e1
    """
    all = sys.modules[f.__module__].__dict__.setdefault("__all__", [])
    if f.__name__ not in all:  # Prevent duplicates if run from an IDE.
        all.append(f.__name__)
    return f


public(public)  # Emulate decorating ourself


class Task(typing.Coroutine[typing.Any, typing.Any, T]):
    """Concurrently executing task.

    This class is not intended for users to directly instantiate.
    Use :func:`cocotb.create_task` to create a Task object,
    or use :func:`cocotb.start_soon` or :func:`cocotb.start` to
    create a Task and schedule it to run.
    """

    _name: str = "Task"  # class name of schedulable task
    _id_count = 0  # used by the scheduler for debug

    def __init__(self, inst):

        if isinstance(inst, collections.abc.Coroutine):
            self._natively_awaitable = True
        elif inspect.isgenerator(inst):
            self._natively_awaitable = False
        elif inspect.iscoroutinefunction(inst):
            raise TypeError(
                "Coroutine function {} should be called prior to being "
                "scheduled.".format(inst)
            )
        elif inspect.isasyncgen(inst):
            raise TypeError(
                "{} is an async generator, not a coroutine. "
                "You likely used the yield keyword instead of await.".format(
                    inst.__qualname__
                )
            )
        else:
            raise TypeError(
                f"{inst} isn't a valid coroutine! Did you forget to use the yield keyword?"
            )
        self._coro = inst
        self._started = False
        self._outcome: outcomes.Outcome = None
        self._trigger: typing.Optional[cocotb.triggers.Trigger] = None
        self._cancelled: typing.Optional[CancelledError] = None

        self._task_id = self._id_count
        type(self)._id_count += 1
        self.__name__ = f"{type(self)._name} {self._task_id}"
        self.__qualname__ = self.__name__

    @lazy_property
    def log(self) -> SimLog:
        # Creating a logger is expensive, only do it if we actually plan to
        # log anything
        return SimLog(f"cocotb.{self.__qualname__}.{self._coro.__qualname__}")

    @property
    def retval(self) -> T:
        """Return the result of the Task.

        If the Task ran to completion, the result is returned.
        If the Task failed with an exception, the exception is re-raised.
        If the Task is not yet complete, a :exc:`RuntimeError` is raised.

        .. deprecated:: 1.7.0
        """
        warnings.warn(
            "Deprecated in favor of the result() method. "
            "Replace `task.retval` with `task.result()`.",
            DeprecationWarning,
            stacklevel=2,
        )
        if self._outcome is None:
            raise RuntimeError("coroutine is not complete")
        return self._outcome.get()

    @property
    def _finished(self) -> bool:
        """``True`` if the Task is finished executing.

        .. deprecated:: 1.7.0
        """
        warnings.warn(
            "Deprecated in favor of the done() method. "
            "Replace `task._finished` with `task.done()`.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._outcome is not None

    def __iter__(self: Self) -> Self:
        # for use in "yield from" statements
        return self

    def __str__(self) -> str:
        return f"<{self.__name__}>"

    def _get_coro_stack(self) -> typing.Any:
        """Get the coroutine callstack of this Task."""
        coro_stack = extract_coro_stack(self._coro)

        # Remove Trigger.__await__() from the stack, as it's not really useful
        if self._natively_awaitable and len(coro_stack):
            if coro_stack[-1].name == "__await__":
                coro_stack.pop()

        return coro_stack

    def __repr__(self) -> str:
        coro_stack = self._get_coro_stack()

        if cocotb.scheduler._current_task is self:
            fmt = "<{name} running coro={coro}()>"
        elif self.done():
            fmt = "<{name} finished coro={coro}() outcome={outcome}>"
        elif self._trigger is not None:
            fmt = "<{name} pending coro={coro}() trigger={trigger}>"
        elif not self._started:
            fmt = "<{name} created coro={coro}()>"
        else:
            fmt = "<{name} adding coro={coro}()>"

        try:
            coro_name = coro_stack[-1].name
        # coro_stack may be empty if:
        # - exhausted generator
        # - finished coroutine
        except IndexError:
            coro_name = self._coro.__name__

        repr_string = fmt.format(
            name=self.__name__,
            coro=coro_name,
            trigger=self._trigger,
            outcome=self._outcome,
        )
        return repr_string

    def _advance(self, outcome: outcomes.Outcome) -> typing.Any:
        """Advance to the next yield in this coroutine.

        Args:
            outcome: The :any:`outcomes.Outcome` object to resume with.

        Returns:
            The object yielded from the coroutine or None if coroutine finished

        """
        try:
            self._started = True
            return outcome.send(self._coro)
        except ReturnValue as e:
            self._outcome = outcomes.Value(e.retval)
        except StopIteration as e:
            self._outcome = outcomes.Value(e.value)
        except BaseException as e:
            self._outcome = outcomes.Error(
                remove_traceback_frames(e, ["_advance", "send"])
            )

    def send(self, value: typing.Any) -> typing.Any:
        return self._coro.send(value)

    def throw(self, exc: BaseException) -> typing.Any:
        return self._coro.throw(exc)

    def close(self) -> None:
        return self._coro.close()

    def kill(self) -> None:
        """Kill a coroutine."""
        if self._outcome is not None:
            # already finished, nothing to kill
            return

        if _debug:
            self.log.debug("kill() called on coroutine")
        # todo: probably better to throw an exception for anyone waiting on the coroutine
        self._outcome = outcomes.Value(None)
        cocotb.scheduler._unschedule(self)

    def join(self) -> cocotb.triggers.Join:
        """Return a trigger that will fire when the wrapped coroutine exits."""
        return cocotb.triggers.Join(self)

    def has_started(self) -> bool:
        """Return ``True`` if the Task has started executing."""
        return self._started

    def cancel(self, msg: typing.Optional[str] = None) -> None:
        """Cancel a Task's further execution.

        When a Task is cancelled, a :exc:`asyncio.CancelledError` is thrown into the Task.
        """
        self._cancelled = CancelledError(msg)
        warnings.warn(
            "Calling this method will cause a CancelledError to be thrown in the "
            "Task sometime in the future.",
            FutureWarning,
            stacklevel=2,
        )
        self.kill()

    def cancelled(self) -> bool:
        """Return ``True`` if the Task was cancelled."""
        return self._cancelled is not None

    def done(self) -> bool:
        """Return ``True`` if the Task has finished executing."""
        return self._outcome is not None or self.cancelled()

    def result(self) -> T:
        """Return the result of the Task.

        If the Task ran to completion, the result is returned.
        If the Task failed with an exception, the exception is re-raised.
        If the Task was cancelled, the CancelledError is re-raised.
        If the coroutine is not yet complete, a :exc:`asyncio.InvalidStateError` is raised.
        """
        if not self.done():
            raise InvalidStateError("result is not yet available")
        elif self.cancelled():
            raise self._cancelled
        else:
            return self._outcome.get()

    def exception(self) -> typing.Optional[BaseException]:
        """Return the exception of the Task.

        If the Task ran to completion, ``None`` is returned.
        If the Task failed with an exception, the exception is returned.
        If the Task was cancelled, the CancelledError is re-raised.
        If the coroutine is not yet complete, a :exc:`asyncio.InvalidStateError` is raised.
        """
        if not self.done():
            raise InvalidStateError("result is not yet available")
        elif self.cancelled():
            raise self._cancelled
        elif isinstance(self._outcome, outcomes.Error):
            return self._outcome.error
        else:
            return None

    def __bool__(self) -> bool:
        """``True`` if Task is not done.

        .. deprecated:: 1.7.0
        """
        warnings.warn(
            "Deprecated in favor of the done() method. "
            "Replace with `not task.done()`.",
            DeprecationWarning,
            stacklevel=2,
        )
        return not self.done()

    def __await__(self) -> typing.Generator[typing.Any, typing.Any, T]:
        # It's tempting to use `return (yield from self._coro)` here,
        # which bypasses the scheduler. Unfortunately, this means that
        # we can't keep track of the result or state of the coroutine,
        # things which we expose in our public API. If you want the
        # efficiency of bypassing the scheduler, remove the `@coroutine`
        # decorator from your `async` functions.

        # Hand the coroutine back to the scheduler trampoline.
        return (yield self)


RunningTask = Task


class RunningCoroutine(Task[T]):
    """
    The result of calling a :any:`cocotb.coroutine` decorated coroutine.

    All this class does is provide some extra attributes.
    """

    def __init__(self, inst, parent):
        super().__init__(inst)
        self._parent = parent
        self.__doc__ = parent._func.__doc__
        self.module = parent._func.__module__
        self.funcname = parent._func.__name__


class RunningTest(RunningCoroutine[T]):
    """
    The result of calling a :class:`cocotb.test` decorated object.

    All this class does is change ``__name__`` to show "Test" instead of "Task".
    """

    _name: str = "Test"

    def __init__(self, inst, parent):
        super().__init__(inst, parent)
        self.__name__ = f"{type(self)._name} {self.funcname}"
        self.__qualname__ = self.__name__


class coroutine:
    """Decorator class that allows us to provide common coroutine mechanisms:

    ``log`` methods will log to ``cocotb.coroutine.name``.

    :meth:`~cocotb.decorators.Task.join` method returns an event which will fire when the coroutine exits.

    Used as ``@cocotb.coroutine``.
    """

    def __init__(self, func):
        self._func = func
        functools.update_wrapper(self, func)

    @lazy_property
    def log(self):
        return SimLog(f"cocotb.coroutine.{self._func.__qualname__}.{id(self)}")

    def __call__(self, *args, **kwargs):
        return RunningCoroutine(self._func(*args, **kwargs), self)

    def __get__(self, obj, owner=None):
        """Permit the decorator to be used on class methods
        and standalone functions"""
        return type(self)(self._func.__get__(obj, owner))

    def __iter__(self):
        return self

    def __str__(self):
        return str(self._func.__qualname__)


@public
class function:
    """Decorator class that allows a function to block.

    This allows a coroutine that consumes simulation time
    to be called by a thread started with :class:`cocotb.external`;
    in other words, to internally block while externally
    appear to yield.
    """

    def __init__(self, func):
        self._coro = cocotb.coroutine(func)

    @lazy_property
    def log(self):
        return SimLog(f"cocotb.function.{self._coro.__qualname__}.{id(self)}")

    def __call__(self, *args, **kwargs):
        return cocotb.scheduler._queue_function(self._coro(*args, **kwargs))

    def __get__(self, obj, owner=None):
        """Permit the decorator to be used on class methods
        and standalone functions"""
        return type(self)(self._coro._func.__get__(obj, owner))


@public
class external:
    """Decorator to apply to an external function to enable calling from cocotb.

    This turns a normal function that isn't a coroutine into a blocking coroutine.
    Currently, this creates a new execution thread for each function that is
    called.
    Scope for this to be streamlined to a queue in future.
    """

    def __init__(self, func):
        self._func = func
        self._log = SimLog(f"cocotb.external.{self._func.__qualname__}.{id(self)}")

    def __call__(self, *args, **kwargs):
        return cocotb.scheduler._run_in_executor(self._func, *args, **kwargs)

    def __get__(self, obj, owner=None):
        """Permit the decorator to be used on class methods
        and standalone functions"""
        return type(self)(self._func.__get__(obj, owner))


class _decorator_helper(type):
    """
    Metaclass that allows a type to be constructed using decorator syntax,
    passing the decorated function as the first argument.

    So:

        @MyClass(construction, args='go here')
        def this_is_passed_as_f(...):
            pass

    ends up calling

        MyClass.__init__(this_is_passed_as_f, construction, args='go here')
    """

    def __call__(cls, *args, **kwargs):
        def decorator(f):
            # fall back to the normal way of constructing an object, now that
            # we have all the arguments
            return type.__call__(cls, f, *args, **kwargs)

        return decorator


@public
class test(coroutine, metaclass=_decorator_helper):
    """
    Decorator to mark a Callable which returns a Coroutine as a test.

    The test decorator provides a test timeout, and allows us to mark tests as skipped
    or expecting errors or failures.
    Tests are evaluated in the order they are defined in a test module.

    Used as ``@cocotb.test(...)``.

    Args:
        timeout_time (numbers.Real or decimal.Decimal, optional):
            Simulation time duration before timeout occurs.

            .. versionadded:: 1.3

            .. note::
                Test timeout is intended for protection against deadlock.
                Users should use :class:`~cocotb.triggers.with_timeout` if they require a
                more general-purpose timeout mechanism.

        timeout_unit (str, optional):
            Units of timeout_time, accepts any units that :class:`~cocotb.triggers.Timer` does.

            .. versionadded:: 1.3

            .. deprecated:: 1.5
                Using ``None`` as the *timeout_unit* argument is deprecated, use ``'step'`` instead.

        expect_fail (bool, optional):
            Don't mark the result as a failure if the test fails.

        expect_error (exception type or tuple of exception types, optional):
            Mark the result as a pass only if one of the exception types is raised in the test.
            This is primarily for cocotb internal regression use for when a simulator error is expected.

            Users are encouraged to use the following idiom instead::

                @cocotb.test()
                async def my_test(dut):
                    try:
                        await thing_that_should_fail()
                    except ExceptionIExpect:
                        pass
                    else:
                        assert False, "Exception did not occur"

            .. versionchanged:: 1.3
                Specific exception types can be expected

            .. deprecated:: 1.5
                Passing a :class:`bool` value is now deprecated.
                Pass a specific :class:`Exception` or a tuple of Exceptions instead.

        skip (bool, optional):
            Don't execute this test as part of the regression. Test can still be run
            manually by setting :make:var:`TESTCASE`.

        stage (int)
            Order tests logically into stages, where multiple tests can share a stage.
            Defaults to 0.
    """

    _id_count = 0  # used by the RegressionManager to sort tests in definition order

    def __init__(
        self,
        f,
        timeout_time=None,
        timeout_unit="step",
        expect_fail=False,
        expect_error=(),
        skip=False,
        stage=0,
    ):

        if timeout_unit is None:
            warnings.warn(
                'Using timeout_unit=None is deprecated, use timeout_unit="step" instead.',
                DeprecationWarning,
                stacklevel=2,
            )
            timeout_unit = "step"  # don't propagate deprecated value
        self._id = self._id_count
        type(self)._id_count += 1

        if timeout_time is not None:
            co = coroutine(f)

            @functools.wraps(f)
            async def f(*args, **kwargs):
                running_co = co(*args, **kwargs)

                try:
                    res = await cocotb.triggers.with_timeout(
                        running_co, self.timeout_time, self.timeout_unit
                    )
                except cocotb.result.SimTimeoutError:
                    running_co.kill()
                    raise
                else:
                    return res

        super().__init__(f)

        self.timeout_time = timeout_time
        self.timeout_unit = timeout_unit
        self.expect_fail = expect_fail
        if isinstance(expect_error, bool):
            warnings.warn(
                "Passing bool values to `except_error` option of `cocotb.test` is deprecated. "
                "Pass a specific Exception type instead",
                DeprecationWarning,
                stacklevel=2,
            )
        if expect_error is True:
            expect_error = (Exception,)
        elif expect_error is False:
            expect_error = ()
        self.expect_error = expect_error
        self.skip = skip
        self.stage = stage
        self.im_test = True  # For auto-regressions
        self.name = self._func.__name__

    def __call__(self, *args, **kwargs):
        inst = self._func(*args, **kwargs)
        coro = RunningTest(inst, self)
        return coro
